from collections import defaultdict
from collections.abc import Sequence

from io import BytesIO

import json
import math
import os
import random
import re

import numpy as np
from tqdm import tqdm

from cloudfiles import CloudFiles

from cloudvolume import CloudVolume
from cloudvolume.lib import min2, Vec, Bbox, mkdir, jsonify
from taskqueue import RegisteredTask, queueable

import fastremap

# from tempfile import NamedTemporaryFile  # used by BigArrayTask

# from backports import lzma               # used by HyperSquareTask
# import blosc                             # used by BigArrayTask
# import h5py                              # used by BigArrayTask

def cache(task, cloudpath):
  layer_path, filename = os.path.split(cloudpath)

  classname = task.__class__.__name__
  lcldir = mkdir(os.path.join('/tmp/', classname))
  lclpath = os.path.join(lcldir, filename)

  if os.path.exists(lclpath):
    with open(lclpath, 'rb') as f:
      filestr = f.read()
  else:
    cf = CloudFiles(layer_path)
    filestr = cf.get(filename)

    with open(lclpath, 'wb') as f:
      f.write(filestr)

  return filestr


class HyperSquareConsensusTask(RegisteredTask):
  """
  Import an Eyewire consensus into neuroglancer by combining
  database information encoded as JSON files with pre-ingested
  Hypersquare.

  The result of the remapping is that all human traced cells should
  be present and identifiable by their Eyewire cells.id number.
  The remaining segments should be the same segid but reencoded
  from 16 to 32 bits such that their high bits encode the
  tasks.segmentation_id according to some mapping that fits
  into 16 bits. It's usually:

    tasks.segmentation_id - min(tasks.segmentation_id for that dataset)

  The consensus map file should be uploaded into the neuroglancer
  data layer directory corresponding to the destination layer being
  processed. The contents of the file are JSON encoded and look like:

  { VOLUMEID: { CELLID: [segids] } }
  """

  def __init__(self, src_path, dest_path, ew_volume_id,
               consensus_map_path, shape, offset):

    super(HyperSquareConsensusTask, self).__init__(
        src_path, dest_path, ew_volume_id,
        consensus_map_path, shape, offset
    )
    self.src_path = src_path
    self.dest_path = dest_path
    self.consensus_map_path = consensus_map_path
    self.shape = Vec(*shape)
    self.offset = Vec(*offset)
    self.ew_volume_id = int(ew_volume_id)

  def execute(self):
    bounds = Bbox(self.offset, self.shape + self.offset)
    srcvol = CloudVolume(self.src_path, fill_missing=True)
    destvol = CloudVolume(self.dest_path)

    consensus = cache(self, self.consensus_map_path).decode('utf8')
    consensus = json.loads(consensus)
    try:
      consensus = consensus[str(self.ew_volume_id)]
    except KeyError:
      print("Black Region", bounds, self.ew_volume_id)
      consensus = {}

    segidmap = self.build_segid_map(consensus, destvol.dtype)

    try:
      image = srcvol[bounds.to_slices()]
    except ValueError:
      print("Skipping", bounds)
      zeroshape = list(bounds.size3()) + [srcvol.num_channels]
      image = np.zeros(shape=zeroshape, dtype=destvol.dtype)

    image = image.astype(destvol.dtype)
    # Merge equivalent segments, non-consensus segments are black
    consensus_image = segidmap[image]

    # Write volume ID to high bits of extended bit width image
    volume_segid_image = image | (self.ew_volume_id << 16)

    # Zero out segid 0 in the high bits so neuroglancer interprets them as empty
    volume_segid_image *= np.logical_not(np.logical_not(image))

    # Final image is consensus keyed by cell ID (C), i.e. 0xCCCCCCCC.
    # Non-empty non-consensus segments are written as:
    # | 16 bit volume_id (V) | 16 bit seg_id (S) |, i.e. 0xVVVVSSSS
    # empties are 0x00000000
    final_image = consensus_image + (consensus_image == 0) * volume_segid_image

    destvol[bounds.to_slices()] = final_image

  def build_segid_map(self, consensus, dtype):
    segidmap = np.zeros(shape=(2 ** 16), dtype=dtype)

    for cellid in consensus:
      for segid in consensus[cellid]:
        segidmap[segid] = int(cellid)

    return segidmap

class WatershedRemapTask(RegisteredTask):
  """
  Take raw watershed output and using a remapping file,
  generate an aggregated segmentation.

  The remap array is a key:value mapping where the
  array index is the key and the value is the contents.

  You can find a script to convert h5 remap files into npy
  files in pipeline/scripts/remap2npy.py

  Required:
      map_path: path to remap file. Must be in npy or npz format.
      src_path: path to watershed layer
      dest_path: path to new layer
      shape: size of volume to remap
      offset: voxel offset into dataset
  """

  def __init__(self, map_path, src_path, dest_path, shape, offset):
    super(WatershedRemapTask, self).__init__(
        map_path, src_path, dest_path, shape, offset)
    self.map_path = map_path
    self.src_path = src_path
    self.dest_path = dest_path
    self.shape = Vec(*shape)
    self.offset = Vec(*offset)

  def execute(self):
    srccv = CloudVolume(self.src_path)
    destcv = CloudVolume(self.dest_path)

    bounds = Bbox(self.offset, self.shape + self.offset)
    bounds = Bbox.clamp(bounds, srccv.bounds)

    remap = self._get_map()
    watershed_data = srccv[bounds.to_slices()]

    # Here's how the remapping works. Numpy has a special
    # indexing that can be used to perform the remap.
    # The remap array is a key:value mapping where the
    # array index is the key and the value is the contents.
    # The watershed_data array contains only data values that
    # are within the length of the remap array.
    #
    # e.g.
    #
    # remap = np.array([1,2,3]) # i.e. 0=>1, 1=>2, 1=>3
    # vals = np.array([0,1,1,1,2,0,2,1,2])
    #
    # remap[vals] # array([1, 2, 2, 2, 3, 1, 3, 2, 3])

    image = remap[watershed_data]
    downsample_and_upload(image, bounds, destcv, self.shape)

  def _get_map(self):
    file = BytesIO(cache(self, self.map_path))
    remap = np.load(file)
    file.close()
    return remap

class MaskAffinitymapTask(RegisteredTask):
    """
    black out the affinitymap regions according to a mask. The mask could be 
    lower resolution in a higher mip level. The affinitymap correspond with 
    zero intensive voxels in the mask will be blacked out. 
    """
    def __init__(self, aff_input_layer_path, aff_output_layer_path, aff_mip, 
                 mask_layer_path, mask_mip, bounds):
        super().__init__(aff_input_layer_path, aff_output_layer_path, aff_mip, 
                         mask_layer_path, mask_mip, bounds)
        self.aff_input_layer_path = aff_input_layer_path 
        self.aff_output_layer_path = aff_output_layer_path 
        self.aff_mip = aff_mip 
        self.mask_layer_path = mask_layer_path
        self.mask_mip = mask_mip 

        self.aff_slices = bounds.to_slices()

    def execute(self):
        self._read_mask()
        self._read_affinity_map()
        self._mask_affinity_map()
        self._upload_affinity_map()

    def _read_affinity_map(self):
        print("download affinity map chunk...")
        if np.all(self.mask==0):
            print("the mask is all black, fill affinitymap with zeros directly")
            sz = (3,) + tuple(s.stop-s.start for s in self.aff_slices) 
            self.aff = np.zeros( sz, dtype='float32' )
            return 

        vol = CloudVolume(self.aff_input_layer_path, bounded=False, fill_missing=True,
                          progress=True, mip=self.aff_mip)
        # the slices did not contain the channel dimension
        self.aff = vol[self.aff_slices[::-1] + (slice(0,3),)]
        self.aff = np.transpose(self.aff)

    def _read_mask(self):
        print("download mask chunk...")
        vol = CloudVolume(self.mask_layer_path, bounded=False, fill_missing=True,
                          progress=True, mip=self.mask_mip)
        self.xyfactor = 2**(self.mask_mip - self.aff_mip)
        # only scale the indices in XY plane 
        self.mask_slices = tuple(slice(a.start//self.xyfactor, a.stop//self.xyfactor) 
                                 for a in self.aff_slices[1:3])
        self.mask_slices = (self.aff_slices[0],) + self.mask_slices 

        # the slices did not contain the channel dimension
        print("mask slices: {}".format(self.mask_slices))
        self.mask = vol[self.mask_slices[::-1]]
        self.mask = np.transpose(self.mask)
        print("shape of mask: {}".format(self.mask.shape))
        self.mask = np.squeeze(self.mask, axis=0)

    def _mask_affinity_map(self):
        if np.all(self.mask):
            print("mask elements are all positive, return directly")
            #return
        if not np.any(self.aff):
            print("affinitymap all black, return directly")
            return 

        print("perform masking ...")
        # use c++ backend 
        # from datatools import mask_affiniy_map 
        # mask_affinity_map(self.aff, self.mask)
        
        assert np.any(self.mask)
        print("upsampling mask ...")
        # upsampling factor in XY plane 
        mask = np.zeros(self.aff.shape[1:], dtype=self.mask.dtype)
        for offset in np.ndindex((self.xyfactor, self.xyfactor)):
            mask[:, np.s_[offset[0]::self.xyfactor], np.s_[offset[1]::self.xyfactor]] = self.mask 

        assert mask.shape == self.aff.shape[1:]
        assert np.any(self.mask)
        np.multiply(self.aff[0,:,:,:], mask, out=self.aff[0,:,:,:])
        np.multiply(self.aff[1,:,:,:], mask, out=self.aff[1,:,:,:])
        np.multiply(self.aff[2,:,:,:], mask, out=self.aff[2,:,:,:])
        assert np.any(self.aff)

    def _upload_affinity_map(self):
        print("upload affinity map chunk...")
        print("output path: {}".format(self.aff_output_layer_path))
        vol = CloudVolume(self.aff_output_layer_path, compress='gzip', 
                          fill_missing=True, bounded=False, autocrop=True, 
                          mip=self.aff_mip, progress=True)
        self.aff = np.transpose(self.aff)
        vol[self.aff_slices[::-1]+(slice(0,3),)] = self.aff 


class InferenceTask(RegisteredTask):
    """
    run inference like ChunkFlow.jl
    1. cutout image using cloudvolume
    2. run inference
    3. crop the margin to make the output aligned with cloud storage backend
    4. upload to cloud storage using cloudvolume

    Note that I always use z,y,x in python, but cloudvolume use x,y,z for indexing.
    So I always do a reverse of slices before indexing.
    """
    def __init__(self, image_layer_path, convnet_path, mask_layer_path, output_layer_path,
            output_offset, output_shape, patch_size, patch_overlap,
            cropping_margin_size, output_key='output', num_output_channels=3, 
                 image_mip=1, output_mip=1, mask_mip=3):
        
        super().__init__(image_layer_path, convnet_path, mask_layer_path, output_layer_path,
                output_offset, output_shape, patch_size, patch_overlap, 
                cropping_margin_size, output_key, num_output_channels, 
                image_mip, output_mip, mask_mip)
        
        output_shape = Vec(*output_shape)
        output_offset = Vec(*output_offset)
        self.image_layer_path = image_layer_path
        self.convnet_path = convnet_path
        self.mask_layer_path = mask_layer_path 
        self.output_layer_path = output_layer_path
        self.output_bounds = Bbox(output_offset, output_shape + output_offset)
        self.patch_size = patch_size
        self.patch_overlap = patch_overlap
        self.cropping_margin_size = cropping_margin_size
        self.output_key = output_key
        self.num_output_channels = num_output_channels
        self.image_mip = image_mip
        self.output_mip = output_mip
        self.mask_mip = mask_mip 
    
    def execute(self):
        self._read_mask()
        # if the mask is black, no need to run inference 
        if np.all(self.mask == 0):
            return 
        self._read_image()
        self._inference()
        self._crop()
        if self.mask: 
            self._mask_output()

        self._upload_output()

    def _read_mask(self):
        if self.mask_layer_path is None or not self.mask_layer_path: 
            print('no mask layer path defined')
            self.mask = None 
            return 
        print("download mask chunk...")
        vol = CloudVolume(self.mask_layer_path, bounded=False, fill_missing=True,
                          progress=True, mip=self.mask_mip)
        self.xyfactor = 2**(self.mask_mip - self.output_mip)
        # only scale the indices in XY plane 
        self.mask_slices = tuple(slice(a.start//self.xyfactor, a.stop//self.xyfactor) 
                                 for a in self.output_bounds.to_slices()[1:3])
        self.mask_slices = (self.output_bounds.to_slices()[0],) + self.mask_slices 

        # the slices did not contain the channel dimension
        print("mask slices: {}".format(self.mask_slices))
        self.mask = vol[self.mask_slices[::-1]]
        self.mask = np.transpose(self.mask)
        print("shape of mask: {}".format(self.mask.shape))
        self.mask = np.squeeze(self.mask, axis=0)

    def _mask_output(self):
        if np.all(self.mask):
            print("mask elements are all positive, return directly")
            #return
        if not np.any(self.output):
            print("output volume is all black, return directly")
            return 

        print("perform masking ...")
        # use c++ backend 
        # from datatools import mask_affiniy_map 
        # mask_affinity_map(self.aff, self.mask)
        
        assert np.any(self.mask)
        print("upsampling mask ...")
        # upsampling factor in XY plane 
        mask = np.zeros(self.output.shape[1:], dtype=self.mask.dtype)
        for offset in np.ndindex((self.xyfactor, self.xyfactor)):
            mask[:, np.s_[offset[0]::self.xyfactor], np.s_[offset[1]::self.xyfactor]] = self.mask 

        assert mask.shape == self.output.shape[1:]
        assert np.any(self.mask)
        np.multiply(self.output[0,:,:,:], mask, out=self.output[0,:,:,:])
        np.multiply(self.output[1,:,:,:], mask, out=self.output[1,:,:,:])
        np.multiply(self.output[2,:,:,:], mask, out=self.output[2,:,:,:])
        assert np.any(self.output)

    def _read_image(self):
        self.vol = CloudVolume(self.image_layer_path, bounded=False, fill_missing=False,
                               progress=True, mip=self.image_mip, parallel=True)
        output_slices = self.output_bounds.to_slices()
        self.input_slices = tuple(slice(s.start - m, s.stop + m) for s, m in
                                  zip(output_slices, self.cropping_margin_size))
        # always reverse the indexes since cloudvolume use x,y,z indexing
        self.image = self.vol[self.input_slices[::-1]]
        # the cutout is fortran ordered, so need to transpose and make it C order
        self.image = np.transpose(self.image)
        self.image = np.ascontiguousarray(self.image)
        assert self.image.shape[0] == 1
        self.image = np.squeeze(self.image, axis=0)

    def _inference(self):
        # prepare for inference
        from chunkflow.block_inference_engine import BlockInferenceEngine
        from chunkflow.frameworks.pznet_patch_inference_engine import PZNetPatchInferenceEngine
        patch_engine = PZNetPatchInferenceEngine(self.convnet_path)
        self.block_inference_engine = BlockInferenceEngine(
            patch_inference_engine=patch_engine,
            patch_size=self.patch_size,
            overlap=self.patch_overlap,
            output_key=self.output_key,
            output_channels=self.num_output_channels)


        # inference engine input is a OffsetArray rather than normal numpy array
        # it is actually a numpy array with global offset
        from chunkflow.offset_array import OffsetArray

        input_offset = tuple(s.start for s in self.input_slices)
        input_chunk = OffsetArray(self.image, global_offset=input_offset)
        self.output = self.block_inference_engine(input_chunk)

    def _crop(self):
        self.output = self.output[:,
                                  self.cropping_margin_size[0] : -self.cropping_margin_size[0],
                                  self.cropping_margin_size[1] : -self.cropping_margin_size[1],
                                  self.cropping_margin_size[2] : -self.cropping_margin_size[2]]

    def _upload_output(self):
        vol = CloudVolume(self.output_layer_path, compress='gzip', fill_missing=True,
                          bounded=False, autocrop=True, mip=self.image_mip, progress=True)
        output_slices = self.output_bounds.to_slices()
        self.output = np.transpose(self.output)
        vol[output_slices[::-1]+(slice(0,self.output.shape[-1]),)] = self.output

# class BigArrayTask(RegisteredTask):
#   def __init__(self, layer_path, chunk_path, chunk_encoding, version):
#     super(BigArrayTask, self).__init__(
#         layer_path, chunk_path, chunk_encoding, version)
#     self.layer_path = layer_path
#     self.chunk_path = chunk_path
#     self.chunk_encoding = chunk_encoding
#     self.version = version

#   def execute(self):
#     self._parse_chunk_path()
#     self._storage = Storage(self.layer_path)
#     self._download_input_chunk()
#     self._upload_chunk()

#   def _parse_chunk_path(self):
#     if self.version == 'zfish_v0/affinities':
#       match = re.match(r'^.*/bigarray/block_(\d+)-(\d+)_(\d+)-(\d+)_(\d+)-(\d+)_1-3.h5$',
#                        self.chunk_path)
#     elif self.version == 'zfish_v0/image' or self.version == 'pinky_v0/image':
#       match = re.match(r'^.*/bigarray/(\d+):(\d+)_(\d+):(\d+)_(\d+):(\d+)$',
#                        self.chunk_path)
#     else:
#       raise NotImplementedError(self.version)

#     (self._xmin, self._xmax,
#      self._ymin, self._ymax,
#      self._zmin, self._zmax) = match.groups()

#     self._xmin = int(self._xmin)
#     self._xmax = int(self._xmax)
#     self._ymin = int(self._ymin)
#     self._ymax = int(self._ymax)
#     self._zmin = int(self._zmin)
#     self._zmax = int(self._zmax)
#     self._filename = self.chunk_path.split('/')[-1]

#   def _download_input_chunk(self):
#     string_data = self._storage.get_file(
#         os.path.join('bigarray', self._filename))
#     if self.version == 'zfish_v0/affinities':
#       self._data = self._decode_hdf5(string_data)
#     elif self.version == 'zfish_v0/image':
#       self._data = self._decode_blosc(string_data, shape=[2048, 2048, 128])
#     elif self.version == 'pinky_v0/image':
#       self._data = self._decode_blosc(string_data, shape=[2048, 2048, 64])
#     else:
#       raise NotImplementedError(self.version)

#   def _decode_blosc(self, string, shape):
#     seeked = blosc.decompress(string[10:])
#     arr = np.fromstring(seeked, dtype=np.uint8).reshape(
#         shape[::-1]).transpose((2, 1, 0))
#     return np.expand_dims(arr, 3)

#   def _decode_hdf5(self, string):
#     with NamedTemporaryFile(delete=False) as tmp:
#       tmp.write(string)
#       tmp.close()
#       with h5py.File(tmp.name, 'r') as h5:
#         return np.transpose(h5['img'][:], axes=(3, 2, 1, 0))

#   def _upload_chunk(self):
#     if self.version == 'zfish_v0/affinities':
#       shape = [313472, 193664, 1280]
#       offset = [14336, 11264, 16384]
#     elif self.version == 'zfish_v0/image':
#       shape = [69632, 34816, 1280]
#       offset = [14336, 12288, 16384]
#     elif self.version == 'pinky_v0/image':
#       shape = [100352, 55296, 1024]
#       offset = [2048, 14336, 16384]
#     else:
#       raise NotImplementedError(self.version)

#     xmin = self._xmin - offset[0] - 1
#     xmax = min(self._xmax - offset[0], shape[0])
#     ymin = self._ymin - offset[1] - 1
#     ymax = min(self._ymax - offset[1], shape[1])
#     zmin = self._zmin - offset[2] - 1
#     zmax = min(self._zmax - offset[2], shape[2])

#     # bigarray chunk has padding to fill the volume
#     chunk = self._data[:xmax-xmin, :ymax-ymin, :zmax-zmin, :]
#     filename = 'build/{:d}-{:d}_{:d}-{:d}_{:d}-{:d}'.format(
#         xmin, xmax, ymin, ymax, zmin, zmax)
#     encoded = self._encode(chunk, self.chunk_encoding)
#     self._storage.put_file(filename, encoded)
#     self._storage.wait_until_queue_empty()

#   def _encode(self, chunk, encoding):
#     if encoding == "jpeg":
#       return chunks.encode_jpeg(chunk)
#     elif encoding == "npz":
#       return chunks.encode_npz(chunk)
#     elif encoding == "npz_uint8":
#       chunk = chunk * 255
#       chunk = chunk.astype(np.uint8)
#       return chunks.encode_npz(chunk)
#     elif encoding == "raw":
#       return chunks.encode_raw(chunk)
#     else:
#       raise NotImplementedError(encoding)


# class HyperSquareTask(RegisteredTask):
#   def __init__(self, bucket_name, dataset_name, layer_name,
#                volume_dir, layer_type, overlap, resolution):

#     self.bucket_name = bucket_name
#     self.dataset_name = dataset_name
#     self.layer_name = layer_name
#     self.volume_dir = volume_dir
#     self.layer_type = layer_type
#     self.overlap = Vec(*overlap)

#     self.resolution = Vec(*resolution)

#     self._volume_cloudpath = 'gs://{}/{}'.format(
#         self.bucket_name, self.volume_dir)
#     self._bucket = None
#     self._metadata = None
#     self._bounds = None

#   def execute(self):
#     client = storage.Client.from_service_account_json(
#         lib.credentials_path(), project=lib.GCLOUD_PROJECT_NAME
#     )
#     self._bucket = client.get_bucket(self.bucket_name)
#     self._metadata = meta = self._download_metadata()

#     self._bounds = Bbox(
#         meta['physical_offset_min'],  # in voxels
#         meta['physical_offset_max']
#     )

#     shape = Vec(*meta['chunk_voxel_dimensions'])
#     shape = Vec(shape.x, shape.y, shape.z, 1)

#     if self.layer_type == 'image':
#       dtype = meta['image_type'].lower()
#       cube = self._materialize_images(shape, dtype)
#     elif self.layer_type == 'segmentation':
#       dtype = meta['segment_id_type'].lower()
#       cube = self._materialize_segmentation(shape, dtype)
#     else:
#       dtype = meta['affinity_type'].lower()
#       return NotImplementedError("Don't know how to get the images for this layer.")

#     self._upload_chunk(cube, dtype)

#   def _download_metadata(self):
#     cloudpath = '{}/metadata.json'.format(self.volume_dir)
#     metadata = self._bucket.get_blob(cloudpath).download_as_string()
#     return json.loads(metadata)

#   def _materialize_segmentation(self, shape, dtype):
#     segmentation_path = '{}/segmentation.lzma'.format(self.volume_dir)
#     seg_blob = self._bucket.get_blob(segmentation_path)
#     return self._decode_lzma(seg_blob.download_as_string(), shape, dtype)

#   def _materialize_images(self, shape, dtype):
#     cloudpaths = ['{}/jpg/{}.jpg'.format(self.volume_dir, i)
#                   for i in xrange(shape.z)]
#     datacube = np.zeros(shape=shape, dtype=np.uint8)  # x,y,z,channels

#     prefix = '{}/jpg/'.format(self.volume_dir)

#     blobs = self._bucket.list_blobs(prefix=prefix)
#     for blob in blobs:
#       z = int(re.findall(r'(\d+)\.jpg', blob.name)[0])
#       imgdata = blob.download_as_string()
#       # Hypersquare images are each situated in the xy plane
#       # so the shape should be (width,height,1)
#       shape = self._bounds.size3()
#       shape.z = 1
#       datacube[:, :, z, :] = chunks.decode_jpeg(imgdata, shape=tuple(shape))

#     return datacube

#   def _decode_lzma(self, string_data, shape, dtype):
#     arr = lzma.decompress(string_data)
#     arr = np.fromstring(arr, dtype=dtype)
#     return arr.reshape(shape[::-1]).T

#   def _upload_chunk(self, datacube, dtype):
#     vol = CloudVolume(self.dataset_name, self.layer_name, mip=0)
#     hov = self.overlap / 2  # half overlap, e.g. 32 -> 16 in e2198
#     img = datacube[hov.x:-hov.x, hov.y:-hov.y,
#                    hov.z:-hov.z, :]  # e.g. 256 -> 224
#     bounds = self._bounds.clone()

#     # the boxes are offset left of zero by half overlap, so no need to
#     # compensate for weird shifts. only upload the non-overlap region.

#     downsample_and_upload(image, bounds, vol, ds_shape=img.shape)
#     vol[bounds.to_slices()] = img


from collections.abc import Sequence

import json
import math
import os
import random
from typing import Optional, Tuple, cast

import numpy as np
from tqdm import tqdm

import fastremap
from mapbuffer import MapBuffer

from cloudfiles import CloudFiles
from taskqueue.registered_task import RegisteredTask

from cloudvolume import CloudVolume
from cloudvolume.exceptions import OutOfBoundsError
from cloudvolume.lib import min2, Vec, Bbox, mkdir, jsonify
from cloudvolume.datasource.precomputed.sharding import ShardingSpecification

from taskqueue import RegisteredTask, queueable
import tinybrain

import igneous.shards
from igneous import downsample_scales
from igneous.types import ShapeType

from .obsolete import (
  HyperSquareConsensusTask, WatershedRemapTask,
  MaskAffinitymapTask, InferenceTask
)

def downsample_and_upload(
    image, bounds, vol, ds_shape,
    mip=0, axis='z', skip_first=False,
    sparse=False, factor=None
  ):
    ds_shape = min2(vol.volume_size, ds_shape[:3])
    underlying_mip = (mip + 1) if (mip + 1) in vol.available_mips else mip
    chunk_size = vol.meta.chunk_size(underlying_mip).astype(np.float32)
    
    if factor is None:
      factor = downsample_scales.axis_to_factor(axis)
    factors = downsample_scales.compute_factors(ds_shape, factor, chunk_size, vol.volume_size)

    if len(factors) == 0:
      print("No factors generated. Image Shape: {}, Downsample Shape: {}, Volume Shape: {}, Bounds: {}".format(
          image.shape, ds_shape, vol.volume_size, bounds)
      )

    vol.mip = mip
    if not skip_first:
      vol[bounds.to_slices()] = image

    if len(factors) == 0:
      return

    num_mips = len(factors)

    mips = []
    if vol.layer_type == 'image':
      mips = tinybrain.downsample_with_averaging(
        image, factors[0], 
        num_mips=num_mips, sparse=sparse
      )
    elif vol.layer_type == 'segmentation':
      mips = tinybrain.downsample_segmentation(
        image, factors[0],
        num_mips=num_mips, sparse=sparse
      )
    else:
      mips = tinybrain.downsample_with_striding(image, factors[0], num_mips=num_mips)

    new_bounds = bounds.clone()

    for factor3 in factors:
      vol.mip += 1
      new_bounds //= factor3
      mipped = mips.pop(0)
      new_bounds.maxpt = new_bounds.minpt + Vec(*mipped.shape[:3])
      vol[new_bounds] = mipped

@queueable
def DeleteTask(layer_path:str, shape, offset, mip=0, num_mips=5):
  """Delete a block of images inside a layer on all mip levels."""
  shape = Vec(*shape)
  offset = Vec(*offset)
  vol = CloudVolume(layer_path, mip=mip, max_redirects=0)

  highres_bbox = Bbox( offset, offset + shape )

  top_mip = min(vol.available_mips[-1], mip + num_mips)

  for mip_i in range(mip, top_mip + 1):
    vol.mip = mip_i
    bbox = vol.bbox_to_mip(highres_bbox, mip, mip_i)
    bbox = bbox.round_to_chunk_size(vol.chunk_size, offset=vol.bounds.minpt)
    bbox = Bbox.clamp(bbox, vol.bounds)

    if bbox.volume() == 0:
      continue

    vol.delete(bbox)

@queueable
def BlackoutTask(
  cloudpath, mip, shape, offset,
  value=0, non_aligned_writes=False
):
  shape = Vec(*shape)
  offset = Vec(*offset)
  vol = CloudVolume(cloudpath, mip, non_aligned_writes=non_aligned_writes)
  bounds = Bbox(offset, shape + offset)
  bounds = Bbox.clamp(bounds, vol.bounds)
  img = np.zeros(bounds.size3(), dtype=vol.dtype) + value
  vol[bounds] = img

@queueable
def TouchTask(cloudpath, mip, shape, offset):
  # This could be made more sophisticated using exists
  vol = CloudVolume(cloudpath, mip, fill_missing=False)
  bounds = Bbox(offset, shape + offset)
  bounds = Bbox.clamp(bounds, vol.bounds)
  image = vol[bounds]

@queueable
def QuantizeTask(
  source_layer_path, dest_layer_path, 
  shape, offset, mip, fill_missing=False
):
  shape = Vec(*shape)
  offset = Vec(*offset)
  srcvol = CloudVolume(source_layer_path, mip=mip,
                       fill_missing=fill_missing)

  bounds = Bbox(offset, shape + offset)
  bounds = Bbox.clamp(bounds, srcvol.bounds)

  image = srcvol[bounds.to_slices()][:, :, :, :1]  # only use x affinity
  image = (image * 255.0).astype(np.uint8)

  destvol = CloudVolume(dest_layer_path, mip=mip)
  downsample_and_upload(image, bounds, destvol, shape, mip=mip, axis='z')

class ContrastNormalizationTask(RegisteredTask):
  """TransferTask + Contrast Correction based on LuminanceLevelsTask output."""
  # translate = change of origin

  def __init__(
    self, src_path, dest_path, levels_path, shape,
    offset, mip, clip_fraction, fill_missing,
    translate, minval, maxval
  ):

    super(ContrastNormalizationTask, self).__init__(
      src_path, dest_path, levels_path, shape, offset,
      mip, clip_fraction, fill_missing, translate,
      minval, maxval
    )
    self.src_path = src_path
    self.dest_path = dest_path
    self.shape = Vec(*shape)
    self.offset = Vec(*offset)
    self.fill_missing = fill_missing
    self.translate = Vec(*translate)
    self.mip = int(mip)
    if isinstance(clip_fraction, Sequence):
      assert len(clip_fraction) == 2
      self.lower_clip_fraction = float(clip_fraction[0])
      self.upper_clip_fraction = float(clip_fraction[1])
    else:
      self.lower_clip_fraction = self.upper_clip_fraction = float(clip_fraction)

    self.minval = minval
    self.maxval = maxval

    self.levels_path = levels_path if levels_path else self.src_path

    assert 0 <= self.lower_clip_fraction <= 1
    assert 0 <= self.upper_clip_fraction <= 1
    assert self.lower_clip_fraction + self.upper_clip_fraction <= 1

  def execute(self):
    srccv = CloudVolume(
        self.src_path, fill_missing=self.fill_missing, mip=self.mip)
    destcv = CloudVolume(
        self.dest_path, fill_missing=self.fill_missing, mip=self.mip)

    bounds = Bbox(self.offset, self.shape[:3] + self.offset)
    bounds = Bbox.clamp(bounds, srccv.bounds)
    image = srccv[bounds.to_slices()].astype(np.float32)

    zlevels = self.fetch_z_levels(bounds)

    nbits = np.dtype(srccv.dtype).itemsize * 8
    maxval = float(2 ** nbits - 1)

    for z in range(bounds.minpt.z, bounds.maxpt.z):
      imagez = z - bounds.minpt.z
      zlevel = zlevels[imagez]
      (lower, upper) = self.find_section_clamping_values(
          zlevel, self.lower_clip_fraction, 1 - self.upper_clip_fraction)
      if lower == upper:
        continue
      img = image[:, :, imagez]
      img = (img - float(lower)) * (maxval / (float(upper) - float(lower)))
      image[:, :, imagez] = img

    image = np.round(image)

    minval = self.minval if self.minval is not None else 0.0
    maxval = self.maxval if self.maxval is not None else maxval

    image = np.clip(image, minval, maxval).astype(destcv.dtype)

    bounds += self.translate
    downsample_and_upload(image, bounds, destcv, self.shape, mip=self.mip)

  def find_section_clamping_values(self, zlevel, lowerfract, upperfract):
    filtered = np.copy(zlevel)

    # remove pure black from frequency counts as
    # it has no information in our images
    filtered[0] = 0

    cdf = np.zeros(shape=(len(filtered),), dtype=np.uint64)
    cdf[0] = filtered[0]
    for i in range(1, len(filtered)):
      cdf[i] = cdf[i - 1] + filtered[i]

    total = cdf[-1]

    if total == 0:
      return (0, 0)

    lower = 0
    for i, val in enumerate(cdf):
      if float(val) / float(total) > lowerfract:
        break
      lower = i

    upper = 0
    for i, val in enumerate(cdf):
      if float(val) / float(total) > upperfract:
        break
      upper = i

    return (lower, upper)

  def fetch_z_levels(self, bounds):
    cf = CloudFiles(self.levels_path)

    levelfilenames = [
      cf.join('levels', f"{self.mip}", f"{z}")
      for z in range(bounds.minpt.z, bounds.maxpt.z)
    ]

    levels = cf.get(levelfilenames)

    errors = [
      level['path'] \
      for level in levels if level['content'] == None
    ]

    if len(errors):
      raise Exception(", ".join(
          errors) + " were not defined. Did you run a LuminanceLevelsTask for these slices?")

    levels = [(
      int(os.path.basename(item['path'])),
      json.loads(item['content'].decode('utf-8'))
    ) for item in levels ]

    levels.sort(key=lambda x: x[0])
    levels = [x[1] for x in levels]
    return [ np.array(x['levels'], dtype=np.uint64) for x in levels ]


class LuminanceLevelsTask(RegisteredTask):
  """Generate a frequency count of luminance values by random sampling. Output to $PATH/levels/$MIP/$Z"""

  def __init__(self, src_path, levels_path, shape, offset, coverage_factor, mip):
    super(LuminanceLevelsTask, self).__init__(
      src_path, levels_path, shape,
      offset, coverage_factor, mip
    )
    self.src_path = src_path
    self.shape = Vec(*shape)
    self.offset = Vec(*offset)
    self.coverage_factor = coverage_factor
    self.mip = int(mip)
    self.levels_path = levels_path

    assert 0 < coverage_factor <= 1, "Coverage Factor must be between 0 and 1"

  def execute(self):
    srccv = CloudVolume(self.src_path, mip=self.mip, fill_missing=True)

    # Accumulate a histogram of the luminance levels
    nbits = np.dtype(srccv.dtype).itemsize * 8
    levels = np.zeros(shape=(2 ** nbits,), dtype=np.uint64)

    bounds = Bbox(self.offset, self.shape[:3] + self.offset)
    bounds = Bbox.clamp(bounds, srccv.bounds)

    bboxes = self.select_bounding_boxes(bounds)
    for bbox in bboxes:
      img2d = srccv[bbox.to_slices()].reshape((bbox.volume()))
      cts = np.bincount(img2d)
      levels[0:len(cts)] += cts.astype(np.uint64)

    covered_area = sum([bbx.volume() for bbx in bboxes])

    bboxes = [(bbox.volume(), bbox.size3()) for bbox in bboxes]
    bboxes.sort(key=lambda x: x[0])
    biggest = bboxes[-1][1]

    output = {
      "levels": levels.tolist(),
      "patch_size": biggest.tolist(),
      "num_patches": len(bboxes),
      "coverage_ratio": covered_area / self.shape.rectVolume(),
    }

    path = self.levels_path if self.levels_path else self.src_path
    path = os.path.join(path, 'levels')

    cf = CloudFiles(path)
    cf.put_json(
      path="{}/{}".format(self.mip, self.offset.z),
      content=output,
      cache_control='no-cache'
    )

  def select_bounding_boxes(self, dataset_bounds):
    # Sample patches until coverage factor is satisfied.
    # Ensure the patches are non-overlapping and random.
    sample_shape = Bbox((0, 0, 0), (2048, 2048, 1))
    area = self.shape.rectVolume()

    total_patches = int(math.ceil(area / sample_shape.volume()))
    N = int(math.ceil(float(total_patches) * self.coverage_factor))

    # Simplification: We are making patch selection against a discrete
    # grid instead of a continuous space. This removes the influence of
    # overlap in a less complex fashion.
    patch_indicies = set()
    while len(patch_indicies) < N:
      ith_patch = random.randint(0, (total_patches - 1))
      patch_indicies.add(ith_patch)

    gridx = int(math.ceil(self.shape.x / sample_shape.size3().x))

    bboxes = []
    for i in patch_indicies:
      patch_start = Vec(i % gridx, i // gridx, 0)
      patch_start *= sample_shape.size3()
      patch_start += self.offset
      bbox = Bbox(patch_start, patch_start + sample_shape.size3())
      bbox = Bbox.clamp(bbox, dataset_bounds)
      bboxes.append(bbox)
    return bboxes

@queueable
def TransferTask(
  src_path, dest_path,
  mip, shape, offset,
  translate=(0,0,0),
  fill_missing=False,
  skip_first=False,
  skip_downsamples=False,
  delete_black_uploads=False,
  background_color=0,
  sparse=False,
  axis='z',
  agglomerate=False,
  timestamp=None,
  compress='gzip',
  factor=None,
):
  """
  Transfer an image to a new location while enabling
  rechunking, translation, reencoding, recompressing,
  and downsampling. For graphene, we can also generate
  proofread segmentation using the agglomerate flag.
  """
  shape = Vec(*shape)
  offset = Vec(*offset)
  fill_missing = bool(fill_missing)
  translate = Vec(*translate)
  delete_black_uploads = bool(delete_black_uploads)
  sparse = bool(sparse)
  skip_first = bool(skip_first)
  skip_downsamples = bool(skip_downsamples)

  src_cv = CloudVolume(
    src_path, fill_missing=fill_missing,
    mip=mip, bounded=False
  )
  dest_cv = CloudVolume(
    dest_path, fill_missing=fill_missing,
    mip=mip, delete_black_uploads=delete_black_uploads,
    background_color=background_color, compress=compress,
  )

  dst_bbox = Bbox(offset, shape + offset)
  dst_bbox = Bbox.clamp(dst_bbox, dest_cv.bounds)

  if (
    skip_downsamples
    and agglomerate == False
    and src_cv.scale == dest_cv.scale
    and src_cv.dtype == dest_cv.dtype
    and np.all(translate == (0,0,0))
  ):
    # most efficient transfer type, just copy
    # files possibly without even decompressing
    src_cv.image.transfer_to(
      dest_path, dst_bbox, mip,
      compress=compress
    )
    return

  src_bbox = dst_bbox - translate
  image = src_cv.download(
    src_bbox, agglomerate=agglomerate, timestamp=timestamp
  )

  if skip_downsamples:
    dest_cv[dst_bbox] = image
  else:
    downsample_and_upload(
      image, dst_bbox, dest_cv,
      shape, mip=mip,
      skip_first=skip_first,
      sparse=sparse, axis=axis,
      factor=factor
    )

@queueable
def DownsampleTask(
  layer_path, mip, shape, offset,
  fill_missing=False, axis='z', sparse=False,
  delete_black_uploads=False, background_color=0,
  dest_path=None, compress="gzip", factor=None
):
  """
  Downsamples a cutout of the volume. By default it performs
  2x2x1 downsamples along the specified axis. The factor argument
  overrrides this functionality.
  """
  if dest_path is None:
    dest_path = layer_path

  return TransferTask(
    layer_path, dest_path,
    mip, shape, offset,
    translate=(0,0,0),
    fill_missing=fill_missing,
    skip_first=True,
    skip_downsamples=False,
    delete_black_uploads=delete_black_uploads,
    background_color=background_color,
    sparse=sparse,
    axis=axis,
    compress=compress,
    factor=factor,
  )

@queueable
def ImageShardTransferTask(
  src_path: str,
  dst_path: str,
  shape: ShapeType,
  offset: ShapeType,
  mip: int = 0,
  fill_missing: bool = False,
  translate: ShapeType = (0, 0, 0),
  agglomerate: bool = False,
  timestamp: Optional[int] = None,
):
  """
  Generates a sharded image volume from
  a preexisting CloudVolume readable data 
  source. Downsamples are not generated.

  The sharded specification can be read here:
  Shard Container: 
  https://github.com/google/neuroglancer/blob/056a3548abffc3c76c93c7a906f1603ce02b5fa3/src/neuroglancer/datasource/precomputed/sharded.md
  Sharded Images:    
  https://github.com/google/neuroglancer/blob/056a3548abffc3c76c93c7a906f1603ce02b5fa3/src/neuroglancer/datasource/precomputed/volume.md#unsharded-chunk-storage
  """
  shape = Vec(*shape)
  offset = Vec(*offset)
  mip = int(mip)
  fill_missing = bool(fill_missing)
  translate = Vec(*translate)

  src_vol = CloudVolume(
    src_path, fill_missing=fill_missing, 
    mip=mip, bounded=False
  )
  dst_vol = CloudVolume(
    dst_path,
    fill_missing=fill_missing,
    mip=mip,
    compress=None,
  )

  dst_bbox = Bbox(offset, offset + shape)
  dst_bbox = Bbox.clamp(dst_bbox, dst_vol.meta.bounds(mip))
  dst_bbox = dst_bbox.expand_to_chunk_size(
    dst_vol.meta.chunk_size(mip), 
    offset=dst_vol.meta.voxel_offset(mip)
  )
  src_bbox = dst_bbox - translate

  img = src_vol.download(
    src_bbox, agglomerate=agglomerate, timestamp=timestamp
  )
  (filename, shard) = dst_vol.image.make_shard(
    img, dst_bbox, mip, progress=False
  )
  del img

  basepath = dst_vol.meta.join(
    dst_vol.cloudpath, dst_vol.meta.key(mip)
  )

  CloudFiles(basepath).put(filename, shard)

@queueable
def ImageShardDownsampleTask(
  src_path: str,
  shape: ShapeType,
  offset: ShapeType,
  mip: int = 0,
  fill_missing: bool = False,
  sparse: bool = False,
  agglomerate: bool = False,
  timestamp: Optional[int] = None,
  factor: ShapeType = (2,2,1)
):
  """
  Generate a single downsample level for a shard.
  Shards are usually hundreds of megabytes to several
  gigabyte of data, so it is usually unrealistic from a
  memory perspective to make more than one mip at a time.
  """
  shape = Vec(*shape)
  offset = Vec(*offset)
  mip = int(mip)
  fill_missing = bool(fill_missing)

  src_vol = CloudVolume(
    src_path, fill_missing=fill_missing, 
    mip=mip, bounded=False, progress=False
  )
  chunk_size = src_vol.meta.chunk_size(mip)

  bbox = Bbox(offset, offset + shape)
  bbox = Bbox.clamp(bbox, src_vol.meta.bounds(mip))
  bbox = bbox.expand_to_chunk_size(
    chunk_size, offset=src_vol.meta.voxel_offset(mip)
  )

  shard_shape = igneous.shards.image_shard_shape_from_spec(
    src_vol.scales[mip + 1]["sharding"], 
    src_vol.meta.volume_size(mip + 1), 
    src_vol.meta.chunk_size(mip + 1)
  )
  upper_offset = offset // Vec(*factor)
  shape_bbox = Bbox(upper_offset, upper_offset + shard_shape)
  shape_bbox = shape_bbox.astype(np.int64)
  shape_bbox = Bbox.clamp(shape_bbox, src_vol.meta.bounds(mip + 1))
  shape_bbox = shape_bbox.expand_to_chunk_size(
    src_vol.meta.chunk_size(mip + 1), 
    offset=src_vol.meta.voxel_offset(mip + 1)
  )

  if shape_bbox.subvoxel():
    return

  shard_shape = list(shape_bbox.size3()) + [ 1 ]

  output_img = np.zeros(shard_shape, dtype=src_vol.dtype, order="F")
  nz = int(math.ceil(bbox.dz / chunk_size.z))

  dsfn = tinybrain.downsample_with_averaging
  if src_vol.layer_type == "segmentation":
    dsfn = tinybrain.downsample_segmentation

  zbox = bbox.clone()
  zbox.maxpt.z = zbox.minpt.z + chunk_size.z
  for z in range(nz):
    img = src_vol.download(
      zbox, agglomerate=agglomerate, timestamp=timestamp
    )
    (ds_img,) = dsfn(img, factor, num_mips=1, sparse=sparse)
    # ds_img[slc] b/c sometimes the size round up in tinybrain
    # makes this too large by one voxel on an axis
    xlim, ylim = tuple(ds_img.shape[:2])
    zs = z * chunk_size.z
    ze = (z+1) * chunk_size.z
    output_img[:xlim,:ylim,zs:ze] = ds_img

    del img
    del ds_img
    zbox.minpt.z += chunk_size.z
    zbox.maxpt.z += chunk_size.z

  (filename, shard) = src_vol.image.make_shard(
    output_img, shape_bbox, (mip + 1), progress=False
  )
  basepath = src_vol.meta.join(
    src_vol.cloudpath, src_vol.meta.key(mip + 1)
  )
  CloudFiles(basepath).put(filename, shard)

@queueable
def CountVoxelsTask(
  cloudpath:str,
  shape:ShapeType,
  offset:ShapeType,
  mip:int = 0,
  fill_missing:bool = False,
  agglomerate:bool = False,
  timestamp:Optional[int] = None
):
  shape = Vec(*shape)
  offset = Vec(*offset)
  mip = int(mip)
  fill_missing = bool(fill_missing)

  cv = CloudVolume(
    cloudpath, fill_missing=fill_missing, 
    mip=mip, bounded=False, progress=False
  )
  bbox = Bbox(offset, offset + shape)
  bbox = Bbox.clamp(bbox, cv.meta.bounds(mip))
  
  labels = cv.download(
    bbox, 
    agglomerate=agglomerate, 
    timestamp=timestamp
  )
  uniq, cts = fastremap.unique(labels, return_counts=True)
  voxel_counts = { str(segid): ct for segid, ct in zip(uniq, cts) }

  cf = CloudFiles(cloudpath)

  cf.put_json(
    cf.join(f'{cv.key}', 'stats', 'voxel_counts', f'{bbox.to_filename()}.json'),
    voxel_counts
  )




from functools import reduce
import math
import multiprocessing as mp
import os
import sys
import time
import webbrowser

import click
from cloudvolume import CloudVolume, Bbox
from cloudvolume.lib import max2
from cloudfiles import CloudFiles
import cloudfiles.paths
import numpy as np
from taskqueue import TaskQueue
from taskqueue.lib import toabs
from taskqueue.paths import get_protocol
from tqdm import tqdm

from igneous import task_creation as tc
from igneous import downsample_scales
from igneous.secrets import LEASE_SECONDS, SQS_REGION_NAME

from igneous_cli.humanbytes import format_bytes

def normalize_path(queuepath):
  if not get_protocol(queuepath):
    return "fq://" + toabs(queuepath)
  return queuepath

def intify(x):
  return None if x is None else int(x)

class Tuple3(click.ParamType):
  """A command line option type consisting of 3 comma-separated integers."""
  name = 'tuple3'
  def convert(self, value, param, ctx):
    if isinstance(value, str):
      try:
        value = tuple(map(int, value.split(',')))
      except ValueError:
        self.fail(f"'{value}' does not contain a comma delimited list of 3 integers.")
      if len(value) != 3:
        self.fail(f"'{value}' does not contain a comma delimited list of 3 integers.")
    return value
  
class Tuple2(click.ParamType):
  """A command line option type consisting of 2 comma-separated integers."""
  name = 'tuple2'
  def convert(self, value, param, ctx):
    if isinstance(value, str):
      try:
        value = tuple(map(int, value.split(',')))
      except ValueError:
        self.fail(f"'{value}' does not contain a comma delimited list of 3 integers.")
      if len(value) != 2:
        self.fail(f"'{value}' does not contain a comma delimited list of 3 integers.")
    return value

class CloudPath(click.ParamType):
  name = "CloudPath"
  def convert(self, value, param, ctx):
    return cloudfiles.paths.normalize(value)

def compute_bounds(path, mip, xrange, yrange, zrange):
  bounds = None
  if xrange or yrange or zrange:
    bounds = CloudVolume(path).meta.bounds(mip)

  if xrange:
    bounds.minpt.x = xrange[0]
    bounds.maxpt.x = xrange[1]
  if yrange:
    bounds.minpt.y = yrange[0]
    bounds.maxpt.y = yrange[1]
  if zrange:
    bounds.minpt.z = zrange[0]
    bounds.maxpt.z = zrange[1]

  return bounds


@click.group()
@click.option("-p", "--parallel", default=1, help="Run with this number of parallel processes. If 0, use number of cores.")
@click.version_option(version="4.14.0")
@click.pass_context
def main(ctx, parallel):
  """
  CLI tool for managing igneous jobs.
  https://github.com/seung-lab/igneous

  Igneous is a tool for producing neuroglancer
  datasets. It scales to hundreds of teravoxels
  or more.

  Select an operation, dataset, and queue and
  tasks will be inserted into the queue. Queues
  can be either SQS or a filesystem directory.

  Then use "igneous execute $queue" to start
  processing that operation.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version. Run "igneous license" for details.
  """
  parallel = int(parallel)
  if parallel == 0:
    parallel = mp.cpu_count()
  ctx.ensure_object(dict)
  ctx.obj["parallel"] = max(min(parallel, mp.cpu_count()), 1)

@main.command()
def license():
  """Prints the license for this library and cli tool."""
  path = os.path.join(os.path.dirname(__file__), 'LICENSE')
  with open(path, 'rt') as f:
    print(f.read())

@main.group("image")
def imagegroup():
  """
  Manipulate image volumes. (subgroup)

  Images are the base datastructure in
  Neuroglancer. This subgroup offers
  methods for downsampling, transfers,
  reencoding, rechunking, sharding,
  and contrast correction.
  """
  pass

@imagegroup.command()
@click.argument("path", type=CloudPath())
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.option('--mip', default=0, help="Build upward from this level of the image pyramid. Default: 0")
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
@click.option('--num-mips', default=5, help="Build this many additional pyramid levels. Each increment increases memory requirements per task 4-8x.  Default: 5")
@click.option('--encoding', default="auto", help="Which image encoding to use. Options: [all] raw, png; [images] jpeg; [segmentations] cseg, compresso; [floats] fpzip, kempressed", show_default=True)
@click.option('--sparse', is_flag=True, default=False, help="Don't count black pixels in mode or average calculations. For images, eliminates edge ghosting in 2x2x2 downsample. For segmentation, prevents small objects from disappearing at high mip levels.")
@click.option('--chunk-size', type=Tuple3(), default=None, help="Chunk size of new layers. e.g. 128,128,64")
@click.option('--compress', default=None, help="Set the image compression scheme. Options: 'gzip', 'br'")
@click.option('--volumetric', is_flag=True, default=False, help="Use 2x2x2 downsampling.")
@click.option('--delete-bg', is_flag=True, default=False, help="Issue a delete instead of uploading a background tile. This is helpful on systems that don't like tiny files.")
@click.option('--bg-color', default=0, help="Determines which color is regarded as background. Default: 0")
@click.option('--sharded', is_flag=True, default=False, help="Generate sharded downsamples which reduces the number of files.")
@click.option('--memory', default=3.5e9, type=int, help="(sharded only) Task memory limit in bytes. Task shape will be chosen to fit and maximize downsamples.", show_default=True)
@click.option('--xrange', type=Tuple2(), default=None, help="If specified, set x-bounds for downsampling in terms of selected mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size (maybe mysterious... use igneous design to investigate). e.g. 0,1024.", show_default=True)
@click.option('--yrange', type=Tuple2(), default=None, help="If specified, set y-bounds for downsampling in terms of selected mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size (maybe mysterious... use igneous design to investigate). e.g. 0,1024", show_default=True)
@click.option('--zrange', type=Tuple2(), default=None, help="If specified, set z-bounds for downsampling in terms of selected mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size (maybe mysterious... use igneous design to investigate). e.g. 0,1", show_default=True)
@click.pass_context
def downsample(
  ctx, path, queue, mip, fill_missing, 
  num_mips, encoding, sparse, 
  chunk_size, compress, volumetric,
  delete_bg, bg_color, sharded, memory,
  xrange, yrange, zrange
):
  """
  Create an image pyramid for grayscale or labeled images.
  By default, we use 2x2x1 downsampling. 

  The levels of the pyramid are called "mips" (from the fake latin 
  "Multum in Parvo" or "many in small"). The base of the pyramid,
  the highest resolution layer, is mip 0. Each level of the pyramid
  is one mip level higher.

  The general strategy is to downsample starting from mip 0. This
  builds several levels. Once that job is complete, pass in the 
  current top mip level of the pyramid. This builds it even taller
  (referred to as "superdownsampling").
  """
  if sharded and num_mips != 1:
    print("igneous: sharded downsamples only support producing one mip at a time.")
    return

  if encoding == "cseg":
    encoding = "compressed_segmentation"
  elif encoding == "auto":
    encoding = None

  factor = (2,2,1)
  if volumetric:
  	factor = (2,2,2)

  bounds = compute_bounds(path, mip, xrange, yrange, zrange)

  if sharded:
    tasks = tc.create_image_shard_downsample_tasks(
      path, mip=mip, fill_missing=fill_missing, 
      sparse=sparse, chunk_size=chunk_size,
      encoding=encoding, memory_target=memory,
      factor=factor, bounds=bounds, bounds_mip=mip
    )
  else:
    tasks = tc.create_downsampling_tasks(
      path, mip=mip, fill_missing=fill_missing, 
      num_mips=num_mips, sparse=sparse, 
      chunk_size=chunk_size, encoding=encoding, 
      delete_black_uploads=delete_bg, 
      background_color=bg_color, 
      compress=compress,
      factor=factor, bounds=bounds,
      bounds_mip=mip
    )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@imagegroup.command()
@click.argument("src", type=CloudPath())
@click.argument("dest", type=CloudPath())
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.option('--mip', default=0, help="Build upward from this level of the image pyramid.", show_default=True)
@click.option('--translate', type=Tuple3(), default=(0, 0, 0), help="Translate the bounding box by X,Y,Z voxels in the new location.")
@click.option('--downsample/--skip-downsample', is_flag=True, default=True, help="Whether or not to produce downsamples from transfer tiles.", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
@click.option('--memory', default=3.5e9, type=int, help="Task memory limit in bytes. Task shape will be chosen to fit and maximize downsamples.", show_default=True)
@click.option('--max-mips', default=5, help="Maximum number of additional pyramid levels.", show_default=True)
@click.option('--encoding', default="auto", help="Which image encoding to use. Options: [all] raw, png; [images] jpeg; [segmentations] cseg, compresso; [floats] fpzip, kempressed", show_default=True)
@click.option('--encoding-level', default=None, help="For some encodings (png level,jpeg quality,fpzip precision) a simple scalar value can adjust the compression efficiency.", show_default=True)
@click.option('--sparse', is_flag=True, default=False, help="Don't count black pixels in mode or average calculations. For images, eliminates edge ghosting in 2x2x2 downsample. For segmentation, prevents small objects from disappearing at high mip levels.", show_default=True)
@click.option('--shape', type=Tuple3(), default=None, help="(overrides --memory) Set the task shape in voxels. This also determines how many downsamples you get. e.g. 2048,2048,64", show_default=True)
@click.option('--chunk-size', type=Tuple3(), default=None, help="Chunk size of destination layer. e.g. 128,128,64", show_default=True)
@click.option('--compress', default="gzip", help="Set the image compression scheme. Options: 'none', 'gzip', 'br'", show_default=True)
@click.option('--volumetric', is_flag=True, default=False, help="Use 2x2x2 downsampling.")
@click.option('--delete-bg', is_flag=True, default=False, help="Issue a delete instead of uploading a background tile. This is helpful on systems that don't like tiny files.")
@click.option('--bg-color', default=0, help="Determines which color is regarded as background.", show_default=True)
@click.option('--sharded', is_flag=True, default=False, help="Generate a sharded dataset which reduces the number of files. Downsamples are not generated.")
@click.option('--dest-voxel-offset', type=Tuple3(), default=None, help="Set the voxel offset for this mip level.")
@click.option('--clean-info', is_flag=True, default=False, help="Scrub info file of mesh and skeleton fields.", show_default=True)
@click.option('--no-src-update', is_flag=True, default=False, help="Don't update the source provenance file with the transfer metadata.", show_default=True)
@click.option('--truncate-scales/--no-truncate-scales', is_flag=True, default=True, help="Truncate the info file scales to the mip specified.", show_default=True)
@click.option('--xrange', type=Tuple2(), default=None, help="If specified, set x-bounds for sampling in terms of selected bounds mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size e.g. 0,1024.", show_default=True)
@click.option('--yrange', type=Tuple2(), default=None, help="If specified, set y-bounds for sampling in terms of selected bounds mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size e.g. 0,1024", show_default=True)
@click.option('--zrange', type=Tuple2(), default=None, help="If specified, set z-bounds for sampling in terms of selected bounds mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size e.g. 0,1", show_default=True)
@click.option('--bounds-mip', default=None, type=int, help="Which mip level are xrange, yrange, and zrange specified in?", show_default=True)
@click.option('--cutout', is_flag=True, default=False, help="If bounds are specified and creating a new volume, restrict the new volume to the specified bounds. Currently only for unsharded transfers.", show_default=True)
@click.pass_context
def xfer(
	ctx, src, dest, queue, translate, 
  downsample, mip, fill_missing, 
  memory, max_mips, shape, sparse, 
  encoding, encoding_level, chunk_size, compress,
  volumetric, delete_bg, bg_color, sharded,
  dest_voxel_offset, clean_info, no_src_update,
  truncate_scales, 
  xrange, yrange, zrange, bounds_mip, cutout
):
  """
  Copy, re-encode, or shard an image layer.

  It is crucial to choose a good task shape. The task
  shape must be a multiple of two of the destination
  image layer chunk size. Too small, and you'll have
  an inefficient transfer. Too big, and you'll run out
  of memory and also have an inefficient transfer.

  Downsamples will by default be automatically calculated
  from whatever material is available. For the default
  2x2x1 downsampling, larger XY dimension is desirable
  compared to Z as more downsamples can be computed for
  each 2x2 increase in the task size.

  Use the --memory flag to automatically compute the
  a reasonable task shape based on your memory limits.
  """
  if encoding == "cseg":
    encoding = "compressed_segmentation"
  elif encoding == "auto":
    encoding = None

  factor = (2,2,1)
  if volumetric:
  	factor = (2,2,2)

  if compress and compress.lower() in ("none", "false"):
    compress = False

  if bounds_mip is None:
    bounds_mip = mip
  bounds = compute_bounds(src, bounds_mip, xrange, yrange, zrange)

  if sharded:
    tasks = tc.create_image_shard_transfer_tasks(
      src, dest,
      chunk_size=chunk_size, fill_missing=fill_missing, mip=mip, 
      dest_voxel_offset=dest_voxel_offset, translate=translate, 
      encoding=encoding, memory_target=memory, clean_info=clean_info,
      encoding_level=encoding_level, truncate_scales=truncate_scales,
      compress=compress, bounds=bounds, bounds_mip=bounds_mip,
      cutout=cutout,
    )
  else:
    tasks = tc.create_transfer_tasks(
      src, dest, 
      chunk_size=chunk_size, fill_missing=fill_missing, 
      dest_voxel_offset=dest_voxel_offset, translate=translate, 
      mip=mip, shape=shape, encoding=encoding, skip_downsamples=(not downsample),
      delete_black_uploads=delete_bg, background_color=bg_color,
      compress=compress, factor=factor, sparse=sparse,
      memory_target=memory, max_mips=max_mips, 
      clean_info=clean_info, no_src_update=no_src_update,
      encoding_level=encoding_level, truncate_scales=truncate_scales,
      bounds=bounds, bounds_mip=bounds_mip, cutout=cutout,
    )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@imagegroup.group("voxels")
def voxelgroup():
  """Compute voxel counts per label."""
  pass

@voxelgroup.command("count")
@click.argument("path", type=CloudPath())
@click.option('--mip', default=0, help="Count this mip level of the image pyramid.", show_default=True)
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.pass_context
def count_voxels(ctx, path, mip, queue):
  """Create voxel counting tasks.

  These tasks are 512x512x512 voxels and result
  in a JSON file that lives at:
  $cloudpath/$KEY/stats/voxel_counts/$BBOX.json
  """
  tasks = tc.create_voxel_counting_tasks(path, mip=mip)
  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@voxelgroup.command("sum")
@click.argument("path", type=CloudPath())
@click.option('--mip', default=0, help="Count this mip level of the image pyramid.", show_default=True)
@click.pass_context
def sum_voxel_counts(ctx, path, mip):
  """Accumulate counts from each task.

  Results are saved in a mapbuffer file:
  $cloudpath/$KEY/stats/voxel_counts.mb
  """
  tc.accumulate_voxel_counts(path, mip=mip)

@imagegroup.group("contrast")
def contrastgroup():
  """Perform contrast correction on the image."""
  pass

@contrastgroup.command()
@click.argument("path", type=CloudPath())
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.option('--mip', default=0, help="Build histogram from this level of the image pyramid.", show_default=True)
@click.option('--coverage', default=0.01, type=float, help="Fraction of the image to sample. Range: [0,1]", show_default=True)
@click.option('--xrange', type=Tuple2(), default=None, help="If specified, set x-bounds for sampling in terms of selected bounds mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size e.g. 0,1024.", show_default=True)
@click.option('--yrange', type=Tuple2(), default=None, help="If specified, set y-bounds for sampling in terms of selected bounds mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size e.g. 0,1024", show_default=True)
@click.option('--zrange', type=Tuple2(), default=None, help="If specified, set z-bounds for sampling in terms of selected bounds mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size e.g. 0,1", show_default=True)
@click.option('--bounds-mip', default=None, type=int, help="Bounds are written in terms of this image pyramid level.", show_default=True)
@click.pass_context
def histogram(
  ctx, path, queue, mip, coverage,
  xrange, yrange, zrange, bounds_mip
):
  """(1) Compute the histogram for each z-slice."""
  if bounds_mip is None:
    bounds_mip = mip
  bounds = compute_bounds(path, bounds_mip, xrange, yrange, zrange)

  tasks = tc.create_luminance_levels_tasks(
    path, 
    levels_path=None, 
    coverage_factor=coverage, 
    mip=mip, 
    bounds_mip=bounds_mip, 
    bounds=bounds,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@contrastgroup.command()
@click.argument("src", type=CloudPath())
@click.argument("dest", type=CloudPath())
@click.option('--shape', default="2048,2048,64", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
@click.option('--translate', type=Tuple3(), default=(0, 0, 0), help="Translate the bounding box by X,Y,Z voxels in the new location.")
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.option('--mip', default=0, help="Apply normalization to this level of the image pyramid.", show_default=True)
@click.option('--clip-fraction', default=0.01, type=float, help="Fraction of histogram on left and right sides to clip. Range: [0,1]", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
@click.option('--minval', default=None, help="Set left side of histogram as this value. (must be in range of datatype)", show_default=True)
@click.option('--maxval', default=None, help="Set right side of histogram as this value. (must be in range of datatype)", show_default=True)
@click.option('--xrange', type=Tuple2(), default=None, help="If specified, set x-bounds for sampling in terms of selected bounds mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size e.g. 0,1024.", show_default=True)
@click.option('--yrange', type=Tuple2(), default=None, help="If specified, set y-bounds for sampling in terms of selected bounds mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size e.g. 0,1024", show_default=True)
@click.option('--zrange', type=Tuple2(), default=None, help="If specified, set z-bounds for sampling in terms of selected bounds mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size e.g. 0,1", show_default=True)
@click.option('--bounds-mip', default=None, type=int, help="Bounds are written in terms of this image pyramid level.", show_default=True)
@click.pass_context
def equalize(
  ctx, src, dest, queue,
  shape, mip, clip_fraction,
  fill_missing, translate,
  minval, maxval,
  xrange, yrange, zrange, bounds_mip
):
  """(2) Apply histogram equalization to z-slices."""
  if bounds_mip is None:
    bounds_mip = mip
  bounds = compute_bounds(path, bounds_mip, xrange, yrange, zrange)

  tasks = tc.create_contrast_normalization_tasks(
    src, dest, levels_path=None,
    shape=shape, mip=mip, clip_fraction=clip_fraction,
    fill_missing=fill_missing, translate=translate,
    minval=minval, maxval=maxval, bounds=bounds,
    bounds_mip=bounds_mip
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@imagegroup.group("ccl")
def cclgroup():
  """
  Perform connected components labeling on the image.

  Result will be a 6-connected labeling of the input
  image. All steps must use the same task shape. If
  a threshold (lte and/or gte) is applied, it must be
  used with the same values in all tasks. This ensures
  that when the CCL is recomputed at each step, the same
  values result.

  Intermediate linkage and relabelign data are saved 
  in PATH/KEY/ccl/{faces,equivalences,relabel}/

  The largest image that can be handled would have 2^64 voxels
  (18 exavoxels, a bit larger than a whole mouse brain).

  Each of the steps are labeled with their sequence number.
  Their order is (1) Generate 3 back faces for each task with 
  1 voxel overlap (so they can be referenced by adjacent tasks)
  (2) Compute linkages between CCL tasks and save the results 
  in a database. (3) Compute a global union find from the linkage 
  data and from that a global relabeling scheme which is saved 
  in the database (4) Apply the relabeling scheme to the image.
  """
  pass

@cclgroup.command("faces")
@click.argument("src", type=CloudPath())
@click.option('--shape', default="512,512,512", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
@click.option('--mip', default=0, help="Apply to this level of the image pyramid.", show_default=True)
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.option('--threshold-gte', default=None, help="Threshold source image using image >= value.", show_default=True)
@click.option('--threshold-lte', default=None, help="Threshold source image using image <= value.", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.", show_default=True)
@click.option('--dust', default=0, help="Delete objects smaller than this number of voxels within a cutout.", show_default=True)
@click.pass_context
def ccl_faces(
  ctx, src, mip, shape, queue,
  threshold_lte, threshold_gte,
  fill_missing, dust
):
  """(1) Generate back face images."""
  tasks = tc.create_ccl_face_tasks(
    src, mip, shape,
    threshold_lte=intify(threshold_lte),
    threshold_gte=intify(threshold_gte),
    fill_missing=fill_missing,
    dust_threshold=dust,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@cclgroup.command("links")
@click.argument("src", type=CloudPath())
@click.option('--shape', default="512,512,512", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
@click.option('--mip', default=0, help="Apply to this level of the image pyramid.", show_default=True)
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.option('--threshold-gte', default=None, help="Threshold source image using image >= value.", show_default=True)
@click.option('--threshold-lte', default=None, help="Threshold source image using image <= value.", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.", show_default=True)
@click.option('--dust', default=0, help="Delete objects smaller than this number of voxels within a cutout.", show_default=True)
@click.pass_context
def ccl_equivalences(
  ctx, src, mip, shape, queue,
  threshold_lte, threshold_gte,
  fill_missing, dust
):
  """(2) Generate links between tasks."""
  tasks = tc.create_ccl_equivalence_tasks(
    src, mip, shape,
    threshold_lte=intify(threshold_lte),
    threshold_gte=intify(threshold_gte),
    fill_missing=fill_missing,
    dust_threshold=dust,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@cclgroup.command("calc-labels")
@click.argument("src", type=CloudPath())
@click.option('--mip', default=0, required=True, help="Apply to this level of the image pyramid.", show_default=True)
@click.option('--shape', default="512,512,512", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
@click.pass_context
def ccl_calc_labels(ctx, src, mip, shape):
  """(3) Compute and serialize a relabeling to the DB."""
  import igneous.tasks.image.ccl
  igneous.tasks.image.ccl.create_relabeling(src, mip, shape)

@cclgroup.command("relabel")
@click.argument("src", type=CloudPath())
@click.argument("dest", type=CloudPath())
@click.option('--shape', default="512,512,512", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
@click.option('--mip', default=0, help="Apply to this level of the image pyramid.", show_default=True)
@click.option('--chunk-size', type=Tuple3(), default=None, help="Chunk size of destination layer. e.g. 128,128,64")
@click.option('--encoding', default="compresso", help="Which image encoding to use. Options: raw, cseg, compresso", show_default=True)
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.option('--threshold-gte', default=None, help="Threshold source image using image >= value.", show_default=True)
@click.option('--threshold-lte', default=None, help="Threshold source image using image <= value.", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.", show_default=True)
@click.option('--dust', default=0, help="Delete objects smaller than this number of voxels within a cutout.", show_default=True)
@click.pass_context
def ccl_relabel(
  ctx, src, dest, 
  shape, mip, chunk_size, 
  encoding, queue,
  threshold_lte, threshold_gte,
  fill_missing, dust
):
  """(4) Finally relabel and write a CCL image."""
  tasks = tc.create_ccl_relabel_tasks(
    src, dest, 
    mip=mip, shape=shape, 
    chunk_size=chunk_size, encoding=encoding,
    threshold_lte=intify(threshold_lte),
    threshold_gte=intify(threshold_gte),
    fill_missing=fill_missing,
    dust_threshold=dust,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@cclgroup.command("clean")
@click.argument("src", type=CloudPath())
@click.option('--mip', default=0, required=True, help="Apply to this level of the image pyramid.", show_default=True)
def ccl_clean(src, mip):
  """(last) Cleans up intermediate files."""
  import igneous.tasks.image.ccl
  igneous.tasks.image.ccl.clean_intermediate_files(src, mip)

@cclgroup.command("auto")
@click.argument("src", type=CloudPath())
@click.argument("dest", type=CloudPath())
@click.option('--shape', default="512,512,512", type=Tuple3(), help="Size of individual tasks in voxels.", show_default=True)
@click.option('--mip', default=0, help="Apply to this level of the image pyramid.", show_default=True)
@click.option('--chunk-size', type=Tuple3(), default=None, help="Chunk size of destination layer. e.g. 128,128,64")
@click.option('--encoding', default="compresso", help="Which image encoding to use. Options: raw, cseg, compresso", show_default=True)
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.option('--clean/--no-clean', default=True, is_flag=True, help="Delete intermediate files on completion.", show_default=True)
@click.option('--threshold-gte', default=None, help="Threshold source image using image >= value.", show_default=True)
@click.option('--threshold-lte', default=None, help="Threshold source image using image <= value.", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.", show_default=True)
@click.option('--dust', default=0, help="Delete objects smaller than this number of voxels within a cutout.", show_default=True)
@click.pass_context
def ccl_auto(
  ctx, src, dest, 
  shape, mip, 
  chunk_size, encoding, 
  queue, clean,
  threshold_lte, threshold_gte,
  fill_missing, dust
):
  """
  For local volumes, execute all steps automatically.
  """
  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  args = (queue, None, LEASE_SECONDS, True, -1, True, False)

  tasks = tc.create_ccl_face_tasks(
    src, mip, shape,
    threshold_lte=intify(threshold_lte),
    threshold_gte=intify(threshold_gte),
    fill_missing=fill_missing,
    dust_threshold=dust,
  )
  tq.insert(tasks, parallel=parallel)
  parallel_execute_helper(parallel, args)

  tasks = tc.create_ccl_equivalence_tasks(
    src, mip, shape,
    threshold_lte=intify(threshold_lte),
    threshold_gte=intify(threshold_gte),
    fill_missing=fill_missing,
    dust_threshold=dust,
  )
  tq.insert(tasks, parallel=parallel)
  parallel_execute_helper(parallel, args)

  import igneous.tasks.image.ccl
  igneous.tasks.image.ccl.create_relabeling(src, mip, shape)

  tasks = tc.create_ccl_relabel_tasks(
    src, dest, 
    mip=mip, shape=shape, 
    chunk_size=chunk_size, encoding=encoding,
    threshold_lte=intify(threshold_lte),
    threshold_gte=intify(threshold_gte),
    fill_missing=fill_missing,
    dust_threshold=dust,
  )
  tq.insert(tasks, parallel=parallel)
  parallel_execute_helper(parallel, args)

  if clean:
    igneous.tasks.image.ccl.clean_intermediate_files(src, mip)

@main.command()
@click.argument("queue", type=str)
@click.option('--aws-region', default=SQS_REGION_NAME, help=f"AWS region in which the SQS queue resides.", show_default=True)
@click.option('--lease-sec', default=LEASE_SECONDS, help=f"Seconds to lease a task for.", type=int, show_default=True)
@click.option('--tally/--no-tally', is_flag=True, default=True, help="Tally completed fq tasks. Does not apply to SQS.", show_default=True)
@click.option('--min-sec', default=-1, help='Execute for at least this many seconds and quit after the last task finishes. Special values: (0) Run at most a single task. (-1) Loop forever (default).', type=float)
@click.option('-q', '--quiet', is_flag=True, default=False, help='Suppress task status messages.', show_default=True)
@click.option('-x', '--exit-on-empty', is_flag=True, default=False, help="Exit immediately when the queue is empty. Not reliable for SQS as measurements are approximate.", show_default=True)
@click.pass_context
def execute(
  ctx, queue, aws_region,
  lease_sec, tally, min_sec,
  exit_on_empty, quiet
):
  """Execute igneous tasks from a queue.

  The queue must be an AWS SQS queue or a FileQueue directory. 
  Examples: (SQS) sqs://my-queue (FileQueue) fq://./my-queue
  (the fq:// is optional).

  See https://github.com/seung-lab/python-task-queue
  """
  parallel = int(ctx.obj.get("parallel", 1))
  args = (queue, aws_region, lease_sec, tally, min_sec, exit_on_empty, quiet)
  parallel_execute_helper(parallel, args)

def parallel_execute_helper(parallel, args):
  if parallel == 1:
    execute_helper(*args)
    return

  pool = mp.Pool(processes=parallel)
  try:
    for _ in range(parallel):
      pool.apply_async(execute_helper, args)
    pool.close()
    pool.join()
  except KeyboardInterrupt:
    print("Interrupted. Exiting.")
    pool.terminate()
    pool.join()

def execute_helper(
  queue, aws_region, lease_sec, 
  tally, min_sec, exit_on_empty,
  quiet
):
  tq = TaskQueue(normalize_path(queue), region_name=aws_region)

  sqs_sec_to_wait = 120
  start_time = time.time()
  last_empty = False

  # Offical Amazon docs state that to determine if a queue is empty,
  # you have to test whether it's consistently empty for several 
  # minutes. So we pick 120 seconds to wait somewhat arbitrarily.
  # https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/confirm-queue-is-empty.html
  def is_empty():
    nonlocal start_time
    nonlocal sqs_sec_to_wait
    nonlocal last_empty

    if tq.path.protocol == "sqs":
      if tq.is_empty():
        if last_empty:
          elapsed_time = time.time() - start_time
          if elapsed_time >= sqs_sec_to_wait:
            return True
          else:
            return False
        else:
          print(f"Queue appearently empty. Waiting {sqs_sec_to_wait} sec. to confirm.")
          start_time = time.time()
          last_empty = True
          return False
      else:
        last_empty = False
        return False
    else:
      return tq.is_empty()

  def stop_after_elapsed_time(tries, elapsed_time):
    if exit_on_empty and is_empty():
      return True

    if min_sec < 0:
      return False
    return min_sec < elapsed_time

  if min_sec != 0:
    tq.poll(
      lease_seconds=lease_sec,
      verbose=(not quiet),
      tally=tally,
      stop_fn=stop_after_elapsed_time,
    )
  else:
    task = tq.lease(seconds=lease_sec)
    task.execute()

@main.group("mesh")
def meshgroup():
  """
  Create 3D meshes from a segmentation. (subgroup)

  Meshing is a two step process of forging then 
  merging. First the meshes are created from a
  regular grid of segmentation cutouts. Second,
  the pieces are glued together.
  """
  pass

@meshgroup.command("xfer")
@click.argument("src", type=CloudPath())
@click.argument("dest", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option("--sharded", is_flag=True, default=False, help="Generate shard fragments instead of outputing mesh fragments.", show_default=True)
@click.option("--dir", "mesh_dir", type=str, default=None, help="Write meshes into this directory instead of the one indicated in the info file.")
@click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks.", show_default=True)
@click.option('--mip', type=int, default=None, help="Manually specify which mip level the images were derived from.", show_default=True)
@click.option('--nlod', type=int, default=0, help="(sharded) Create additional levels of detail (lods).", show_default=True)
@click.pass_context
def mesh_xfer(
  ctx, src, dest, queue,
  sharded, mesh_dir, magnitude,
  mip, nlod
):
  """
  Transfer meshes to another location.
  Enables conversion of unsharded to sharded
  as well.
  """
  cv_src = CloudVolume(src)

  if not cv_src.mesh.meta.is_sharded() and sharded:
    tasks = tc.create_sharded_multires_mesh_from_unsharded_tasks(
      src, dest,
      num_lod=nlod, 
      mesh_dir=mesh_dir, 
      mip=mip,
    )
  else:
    tasks = tc.create_xfer_meshes_tasks(
      src, dest, 
      mesh_dir=mesh_dir, 
      magnitude=magnitude,
    )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@meshgroup.command("forge")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--mip', default=0, help="Perform meshing using this level of the image pyramid.", show_default=True)
@click.option('--shape', type=Tuple3(), default=(448, 448, 448), help="Set the task shape in voxels.", show_default=True)
@click.option('--simplify/--skip-simplify', is_flag=True, default=True, help="Enable mesh simplification.", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
@click.option('--max-error', default=40, help="Maximum simplification error in physical units.", show_default=True)
@click.option('--dust-threshold', default=None, help="Skip meshing objects smaller than this number of voxels within a cutout. No default limit. Typical value: 1000.", type=int)
@click.option('--dust-global/--dust-local', is_flag=True, default=False, help="Use global voxel counts for the dust threshold (when >0). To use this feature you must first compute the global voxel counts using the 'igneous image voxels' command.", show_default=True)
@click.option('--dir', default=None, help="Write meshes into this directory instead of the one indicated in the info file.")
@click.option('--compress', default="gzip", help="Set the image compression scheme. Options: 'none', 'gzip', 'br'", show_default=True)
@click.option('--spatial-index/--skip-spatial-index', is_flag=True, default=True, help="Create the spatial index.", show_default=True)
@click.option('--sharded', is_flag=True, default=False, help="Generate shard fragments instead of outputing mesh fragments.", show_default=True)
@click.option('--closed-edge/--open-edge', is_flag=True, default=True, help="Whether meshes are closed on the side that contacts the dataset boundary.", show_default=True)
@click.pass_context
def mesh_forge(
  ctx, path, queue, mip, shape, 
  simplify, fill_missing, max_error, 
  dust_threshold, dir, compress, 
  spatial_index, sharded, closed_edge,
  dust_global
):
  """
  (1) Synthesize meshes from segmentation cutouts.

  A large labeled image is divided into a regular
  grid. zmesh is applied to grid point, which performs
  marching cubes and a quadratic mesh simplifier.

  Note that using task shapes with axes less than
  or equal to 511x1023x511 (don't ask) will be more
  memory efficient as it can use a 32-bit mesher.

  zmesh is used: https://github.com/seung-lab/zmesh
  """
  if compress.lower() == "none":
    compress = False

  tasks = tc.create_meshing_tasks(
    path, mip, shape, 
    simplification=simplify, max_simplification_error=max_error,
    mesh_dir=dir, cdn_cache=False, dust_threshold=dust_threshold,
    object_ids=None, progress=False, fill_missing=fill_missing,
    encoding='precomputed', spatial_index=spatial_index, 
    sharded=sharded, compress=compress, closed_dataset_edges=closed_edge,
    dust_global=dust_global
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@meshgroup.command("merge")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks. Default: 2 (100 tasks)")
@click.option('--nlod', default=0, help="(multires) How many extra levels of detail to create.", show_default=True)
@click.option('--vqb', default=16, help="(multires) Vertex quantization bits for stored model representation. 10 or 16 only.", show_default=True)
@click.option('--min-chunk-size', type=Tuple3(), default="256,256,256",  help="(multires) Sets the minimum chunk size of the highest resolution mesh fragment.", show_default=True)
@click.option('--dir', default=None, help="Write manifests into this directory instead of the one indicated in the info file.")
@click.pass_context
def mesh_merge(ctx, path, queue, magnitude, nlod, vqb, dir, min_chunk_size):
  """
  (2) Merge the mesh pieces produced from the forging step.

  The per-cutout mesh fragments are then assembled and
  merged. However, this process occurs by compiling 
  a list of fragment files and uploading a "mesh manifest"
  file that is an index for locating the fragments.
  """
  if nlod > 0:
    tasks = tc.create_unsharded_multires_mesh_tasks(
      path, num_lod=nlod, 
      magnitude=magnitude, mesh_dir=dir,
      vertex_quantization_bits=vqb,
      min_chunk_size=min_chunk_size,
    )
  else:
    tasks = tc.create_mesh_manifest_tasks(
      path, magnitude=magnitude, mesh_dir=dir
    )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@meshgroup.command("merge-sharded")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--nlod', default=1, help="Number of levels of detail to create.", type=int, show_default=True)
@click.option('--vqb', default=16, help="Vertex quantization bits. Can be 10 or 16.", type=int, show_default=True)
@click.option('--compress-level', default=7, help="Draco compression level.", type=int, show_default=True)
@click.option('--shard-index-bytes', default=2**13, help="Size in bytes to make the shard index.", type=int, show_default=True)
@click.option('--minishard-index-bytes', default=2**15, help="Size in bytes to make the minishard index.", type=int, show_default=True)
@click.option('--min-shards', default=1, help="Minimum number of shards to generate. Excess shards can make it easier to parallelize the merge process.", type=int, show_default=True)
@click.option('--max-labels-per-shard', default=1000, help="Maximum average number of labels per a shard.", type=int, show_default=True)
@click.option('--minishard-index-encoding', default="gzip", help="Minishard indices can be compressed. gzip or raw.", show_default=True)
@click.option('--spatial-index-db', default=None, help="CloudVolume generated SQL database for spatial index.", show_default=True)
@click.option('--min-chunk-size', type=Tuple3(), default="256,256,256",  help="(multires) Sets the minimum chunk size of the highest resolution mesh fragment.", show_default=True)
@click.pass_context
def mesh_sharded_merge(
  ctx, path, queue, 
  nlod, vqb, compress_level,
  shard_index_bytes, minishard_index_bytes, min_shards,
  max_labels_per_shard, minishard_index_encoding, 
  spatial_index_db, min_chunk_size
):
  """
  (2) Postprocess fragments into finished sharded multires meshes.

  Only use this command if you used the --sharded flag
  during the forging step. Some reasonable defaults
  are selected for a dataset with a few million labels,
  but for smaller or larger datasets they may not be
  appropriate.

  The shard and minishard index default sizes are set to
  accomodate efficient access for a 100 Mbps connection.
  """
  tasks = tc.create_sharded_multires_mesh_tasks(
    path, 
    num_lod=nlod,
    draco_compression_level=compress_level,
    vertex_quantization_bits=vqb,
    shard_index_bytes=shard_index_bytes,
    minishard_index_bytes=minishard_index_bytes,
    minishard_index_encoding=minishard_index_encoding,
    min_shards=min_shards,
    spatial_index_db=spatial_index_db,
    min_chunk_size=min_chunk_size,
    max_labels_per_shard=max_labels_per_shard,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@meshgroup.command("rm")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks. Default: 2 (100 tasks)")
@click.option('--dir', 'mesh_dir', default=None, help="Target this directory instead of the one indicated in the info file.")
@click.pass_context
def mesh_rm(ctx, path, queue, magnitude, mesh_dir):
  """
  Delete mesh files.
  """
  tasks = tc.create_mesh_deletion_tasks(
    path, magnitude=magnitude, mesh_dir=mesh_dir
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@meshgroup.group("spatial-index")
def spatialindexgroup():
  """
  (subgroup) Create or download mesh spatial indices.
  """
  pass

@spatialindexgroup.command("create")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--shape', default="448,448,448", type=Tuple3(), help="Shape in voxels of each indexing task.", show_default=True)
@click.option('--mip', default=0, help="Perform indexing using this level of the image pyramid.", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.", show_default=True)
@click.pass_context
def mesh_spatial_index_create(ctx, path, queue, shape, mip, fill_missing):
  """
  Create a spatial index on a pre-existing mesh.

  Sometimes datasets were meshes without a
  spatial index or need it to be updated.
  This function provides a more efficient
  way to accomplish that than remeshing.
  """
  tasks = tc.create_spatial_index_mesh_tasks(
    cloudpath=path,
    shape=shape,
    mip=mip, 
    fill_missing=fill_missing,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@spatialindexgroup.command("db")
@click.argument("path", type=CloudPath())
@click.argument("database")
@click.option('--progress', is_flag=True, default=False, help="Show progress bars.", show_default=True)
@click.option('--allow-missing', is_flag=True, default=False, help="Allow missing index files.", show_default=True)
@click.pass_context
def mesh_spatial_index_download(ctx, path, database, progress, allow_missing):
  """
  Download the mesh spatial index into a database.

  sqlite paths: sqlite://filename.db (prefix optional)
  mysql paths: mysql://{user}:{pwd}@{host}/{database}
  """
  parallel = int(ctx.obj.get("parallel", 1))
  cv = CloudVolume(path)
  cv.mesh.spatial_index.to_sql(
    database, 
    progress=progress,
    allow_missing=allow_missing, 
    parallel=parallel,
  )

@main.group("skeleton")
def skeletongroup():
  """
  Create skeletons from a segmentation.

  Skeletonizing is a two step process of forging 
  then merging. First the skeletons are created
  from a regular grid of segmentation cutouts.
  Second, the pieces are postprocessed and glued
  together.
  """
  pass

@skeletongroup.command("forge")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--mip', default=0, help="Perform skeletonizing using this level of the image pyramid.", show_default=True)
@click.option('--shape', type=Tuple3(), default=(512, 512, 512), help="Set the task shape in voxels.", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.", show_default=True)
@click.option('--fix-branching', is_flag=True, default=True, help="Trades speed for quality of branching at forks. Default: True")
@click.option('--fix-borders', is_flag=True, default=True, help="Allows trivial merging of single voxel overlap tasks. Only switch off for datasets that fit in a single task.", show_default=True)
@click.option('--fix-avocados', is_flag=True, default=False, help="Fixes somata where nuclei and cytoplasm have separate segmentations.", show_default=True)
@click.option('--fill-holes', is_flag=True, default=False, help="Preprocess each cutout to eliminate background holes and holes caused by entirely contained inclusions. Warning: May remove labels that are considered inclusions.", show_default=True)
@click.option('--dust-threshold', default=1000, help="Skip skeletonizing objects smaller than this number of voxels within a cutout.", type=int, show_default=True)
@click.option('--dust-global/--dust-local', is_flag=True, default=False, help="Use global voxel counts for the dust threshold (when >0). To use this feature you must first compute the global voxel counts using the 'igneous image voxels' command.", show_default=True)
@click.option('--spatial-index/--skip-spatial-index', is_flag=True, default=True, help="Create the spatial index.", show_default=True)
@click.option('--scale', default=4, help="Multiplies invalidation radius by distance from boundary.", type=float, show_default=True)
@click.option('--const', default=10, help="Adds constant amount to invalidation radius in physical units.", type=float, show_default=True)
@click.option('--soma-detect', default=1100, help="Consider objects with peak distances to boundary larger than this soma candidates. Physical units.", type=float, show_default=True)
@click.option('--soma-accept', default=3500, help="Accept soma candidates over this threshold and perform a one-time spherical invalidation around their peak value. Physical units.", type=float, show_default=True)
@click.option('--soma-scale', default=1.0, help="Scale factor for soma invalidation.", type=float, show_default=True)
@click.option('--soma-const', default=300, help="Const factor for soma invalidation.", type=float, show_default=True)
@click.option('--max-paths', default=None, help="Abort skeletonizing an object after this many paths have been traced.", type=float)
@click.option('--sharded', is_flag=True, default=False, help="Generate shard fragments instead of outputing skeleton fragments.", show_default=True)
@click.pass_context
def skeleton_forge(
  ctx, path, queue, mip, shape, 
  fill_missing, dust_threshold, dust_global, spatial_index,
  fix_branching, fix_borders, fix_avocados, 
  fill_holes, scale, const, soma_detect, soma_accept,
  soma_scale, soma_const, max_paths, sharded
):
  """
  (1) Synthesize skeletons from segmentation cutouts.

  A large labeled image is divided into a regular
  grid. Kimimaro is applied to grid point, which performs
  a TEASAR based skeletonization.

  You can read more about the parameters here:
  https://github.com/seung-lab/kimimaro

  Tutorials are located here:

  - https://github.com/seung-lab/kimimaro/wiki/A-Pictorial-Guide-to-TEASAR-Skeletonization

  - https://github.com/seung-lab/kimimaro/wiki/Intuition-for-Setting-Parameters-const-and-scale

  A guide to how much this might cost is located here:

  - https://github.com/seung-lab/kimimaro/wiki/The-Economics:-Skeletons-for-the-People
  """
  teasar_params = {
    'scale': scale,
    'const': const, # physical units
    'pdrf_exponent': 4,
    'pdrf_scale': 100000,
    'soma_detection_threshold': soma_detect, # physical units
    'soma_acceptance_threshold': soma_accept, # physical units
    'soma_invalidation_scale': soma_scale,
    'soma_invalidation_const': soma_const, # physical units
    'max_paths': max_paths, # default None
  }

  tasks = tc.create_skeletonizing_tasks(
    path, mip, shape, 
    teasar_params=teasar_params, 
    fix_branching=fix_branching, fix_borders=fix_borders, 
    fix_avocados=fix_avocados, fill_holes=fill_holes,
    dust_threshold=dust_threshold, progress=False,
    parallel=1, fill_missing=fill_missing, 
    sharded=sharded, spatial_index=spatial_index,
    dust_global=dust_global,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)


@skeletongroup.command("merge")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--min-cable-length', default=1000, help="Skip objects smaller than this physical path length. Default: 1000 nm", type=float)
@click.option('--max-cable-length', default=None, help="Skip objects larger than this physical path length. Default: no limit", type=float)
@click.option('--tick-threshold', default=0, help="Remove small \"ticks\", or branches from the main skeleton one at a time from smallest to largest. Branches larger than this are preserved. Default: no elimination", type=float)
@click.option('--delete-fragments', is_flag=True, default=False, help="Delete the skeleton fragments from the first pass after upload is complete.")
@click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks. Default: 2 (100 tasks)", type=int)
@click.pass_context
def skeleton_merge(
  ctx, path, queue, 
  min_cable_length, max_cable_length, 
  tick_threshold, delete_fragments, 
  magnitude
):
  """
  (2) Postprocess fragments into finished skeletons.
  """
  tasks = tc.create_unsharded_skeleton_merge_tasks(
    path, 
    magnitude=magnitude, 
    dust_threshold=min_cable_length,
    max_cable_length=max_cable_length,
    tick_threshold=tick_threshold,
    delete_fragments=delete_fragments,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@skeletongroup.command("merge-sharded")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--min-cable-length', default=1000, help="Skip objects smaller than this physical path length.", type=float, show_default=True)
@click.option('--max-cable-length', default=None, help="Skip objects larger than this physical path length. Default: no limit", type=float)
@click.option('--tick-threshold', default=0, help="Remove small \"ticks\", or branches from the main skeleton one at a time from smallest to largest. Branches larger than this are preserved. Default: no elimination", type=float)
@click.option('--shard-index-bytes', default=2**13, help="Size in bytes to make the shard index.", type=int, show_default=True)
@click.option('--minishard-index-bytes', default=2**15, help="Size in bytes to make the minishard index.", type=int, show_default=True)
@click.option('--max-labels-per-shard', default=2000, help="Maximum average number of labels per a shard.", type=int, show_default=True)
@click.option('--min-shards', default=1, help="Minimum number of shards to generate. Excess shards can make it easier to parallelize the merge process.", type=int, show_default=True)
@click.option('--minishard-index-encoding', default="gzip", help="Minishard indices can be compressed. gzip or raw. Default: gzip")
@click.option('--data-encoding', default="gzip", help="Shard data can be compressed. gzip or raw. Default: gzip")
@click.option('--spatial-index-db', default=None, help="CloudVolume generated SQL database for spatial index.", show_default=True)
@click.pass_context
def skeleton_sharded_merge(
  ctx, path, queue, 
  min_cable_length, max_cable_length, 
  tick_threshold, 
  shard_index_bytes, minishard_index_bytes, 
  max_labels_per_shard, min_shards,
  minishard_index_encoding, data_encoding,
  spatial_index_db
):
  """
  (2) Postprocess fragments into finished skeletons.

  Only use this command if you used the --sharded flag
  during the forging step. Some reasonable defaults
  are selected for a dataset with a few million labels,
  but for smaller or larger datasets they may not be
  appropriate.

  The shard and minishard index default sizes are set to
  accomodate efficient access for a 100 Mbps connection.
  """
  tasks = tc.create_sharded_skeleton_merge_tasks(
    path, 
    dust_threshold=min_cable_length,
    max_cable_length=max_cable_length,
    tick_threshold=tick_threshold,
    shard_index_bytes=shard_index_bytes,
    minishard_index_bytes=minishard_index_bytes,
    min_shards=min_shards,
    minishard_index_encoding=minishard_index_encoding, 
    data_encoding=data_encoding,
    spatial_index_db=spatial_index_db,
    max_labels_per_shard=max_labels_per_shard,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@imagegroup.command("rm")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--mip', default=0, help="Which mip level to start deleting from. Default: 0")
@click.option('--num-mips', default=5, help="The number of mip levels to delete at once. Default: 5")
@click.option('--shape', default=None, help="The size of each deletion task as a comma delimited list. Must be a multiple of the chunk size.", type=Tuple3())
@click.pass_context
def delete_images(
  ctx, path, queue, 
  mip, num_mips, shape
):
  """
  Delete the image layer of a dataset.
  """
  tasks = tc.create_deletion_tasks(
    path, mip, num_mips=num_mips, shape=shape
  )
  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@skeletongroup.command("rm")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks. Default: 2 (100 tasks)")
@click.option('--dir', 'skel_dir', default=None, help="Target this directory instead of the one indicated in the info file.")
@click.pass_context
def skeleton_rm(ctx, path, queue, magnitude, skel_dir):
  """
  Delete skeleton files.
  """
  tasks = tc.create_skeleton_deletion_tasks(
    path, magnitude=magnitude, skel_dir=skel_dir
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)


@skeletongroup.command("xfer")
@click.argument("src", type=CloudPath())
@click.argument("dest", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option("--sharded", is_flag=True, default=False, help="Generate shard fragments instead of outputing mesh fragments.", show_default=True)
@click.option("--dir", "skel_dir", type=str, default=None, help="Write skeletons into this directory instead of the one indicated in the info file.")
@click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks.", show_default=True)
@click.pass_context
def skel_xfer(
  ctx, src, dest, queue,
  sharded, skel_dir, magnitude
):
  """
  Transfer skeletons to another location.
  Enables conversion of unsharded to sharded
  as well.
  """
  cv_src = CloudVolume(src)

  if not cv_src.skeleton.meta.is_sharded() and sharded:
    tasks = tc.create_sharded_skeletons_from_unsharded_tasks(
      src, dest, skel_dir=skel_dir,
    )
  else:
    tasks = tc.create_xfer_skeleton_tasks(
      src, dest,
      skel_dir=skel_dir, 
      magnitude=magnitude,
    )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@skeletongroup.group("spatial-index")
def spatialindexgroupskel():
  """
  (subgroup) Create or download mesh spatial indices.
  """
  pass

@spatialindexgroupskel.command("create")
@click.argument("path", type=CloudPath())
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--shape', default="512,512,512", type=Tuple3(), help="Shape in voxels of each indexing task.", show_default=True)
@click.option('--mip', default=0, help="Perform indexing using this level of the image pyramid.", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.", show_default=True)
@click.pass_context
def skel_spatial_index_create(ctx, path, queue, shape, mip, fill_missing):
  """
  Create a spatial index on pre-existing skeletons.

  Sometimes datasets were skeletonized without a
  spatial index or need it to be updated.
  This function provides a more efficient
  way to accomplish that than reskeletonizing.
  """
  tasks = tc.create_spatial_index_skeleton_tasks(
    cloudpath=path,
    shape=shape,
    mip=mip, 
    fill_missing=fill_missing,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@spatialindexgroupskel.command("db")
@click.argument("path", type=CloudPath())
@click.argument("database")
@click.option('--progress', is_flag=True, default=False, help="Show progress bars.", show_default=True)
@click.option('--allow-missing', is_flag=True, default=False, help="Allow missing index files.", show_default=True)
@click.pass_context
def skel_spatial_index_download(ctx, path, database, progress, allow_missing):
  """
  Download the skeleton spatial index into a database.

  sqlite paths: sqlite://filename.db (prefix optional)
  mysql paths: mysql://{user}:{pwd}@{host}/{database}
  """
  parallel = int(ctx.obj.get("parallel", 1))
  cv = CloudVolume(path)
  cv.skeleton.spatial_index.to_sql(
    database, 
    progress=progress,
    allow_missing=allow_missing, 
    parallel=parallel,
  )

@main.group("design")
def designgroup():
  """
  Tools to aid the design of tasks or neuroglancer layers.
  """
  pass

@designgroup.command("bounds")
@click.argument("path", type=CloudPath())
@click.option('--mip', default=0, help="Select level of the image pyramid.", show_default=True)
def dsbounds(path, mip):
  """
  Detects the volume bounds and chunk size for
  an unsharded image volume. Useful when there
  is a corrupted info file.
  """
  cv = CloudVolume(path, mip=mip)
  cf = CloudFiles(path)

  bboxes = []
  for filename in tqdm(cf.list(prefix=cv.key), desc="Computing Bounds"):
    bboxes.append( Bbox.from_filename(filename) )

  bounds = Bbox.expand(*bboxes)
  chunk_size = list(reduce(max2, map(lambda bbox: bbox.size3(), bboxes)))

  print(f"Bounds: {bounds}")
  print(f"Volume: {list(bounds.size3())}")
  print(f"Chunks: {chunk_size}")

@designgroup.command("ds-memory")
@click.argument("path", type=CloudPath())
@click.argument("memory_bytes", type=float)
@click.option('--mip', default=0, help="Select level of the image pyramid.", show_default=True)
@click.option('--factor', default="2,2,1", type=Tuple3(), help="Downsample factor to use.", show_default=True)
@click.option('--verbose', is_flag=True, help="Print some additional information.")
@click.option('--max-mips', default=5, help="Maximum downsamples to generate from this shape.", show_default=True)
def dsmemory(path, memory_bytes, mip, factor, verbose, max_mips):
  """
  Compute the task shape that maximizes the number of
  downsamples for a given amount of memory.
  """
  cv = CloudVolume(path, mip=mip)

  data_width = np.dtype(cv.dtype).itemsize
  cx, cy, cz = cv.chunk_size
  memory_bytes = int(memory_bytes)

  shape = downsample_scales.downsample_shape_from_memory_target(
    data_width, 
    cx, cy, cz, factor, 
    memory_bytes, max_mips
  )

  num_downsamples = int(math.log2(max(shape / cv.chunk_size)))

  mem_used = memory_used(data_width, shape, factor)

  shape = [ str(x) for x in shape ]
  shape = ",".join(shape)

  if verbose:
    print(f"Data Width: {data_width} B")
    print(f"Factor: {factor}")
    print(f"Chunk Size: {cx}, {cy}, {cz} voxels")
    print(f"Memory Limit: {format_bytes(memory_bytes, True)}")
    print("-----")
    print(f"Optimized Shape: {shape}")
    print(f"Downsamples: {num_downsamples}")
    print(f"Memory Used*: {format_bytes(mem_used, True)}")
    print(
      "\n"
      "* memory used is for retaining the image "
      "and all downsamples.\nAdditional costs may be incurred "
      "from processing."
    )
  else:
    print(shape)

@designgroup.command("ds-shape")
@click.argument("path", type=CloudPath())
@click.argument("shape", type=Tuple3())
@click.option('--mip', default=0, help="Select level of the image pyramid.", show_default=True)
@click.option('--factor', default="2,2,1", type=Tuple3(), help="Downsample factor to use.", show_default=True)
def dsshape(path, shape, mip, factor):
  """
  Compute the approximate memory usage for a
  given downsample task shape.
  """
  cv = CloudVolume(path, mip=mip)
  data_width = np.dtype(cv.dtype).itemsize
  memory_bytes = memory_used(data_width, shape, factor)  
  print(format_bytes(memory_bytes, True))

def memory_used(data_width, shape, factor): 
  memory_bytes = data_width * shape[0] * shape[1] * shape[2]

  if factor not in ((1,1,1), (2,2,1), (2,1,2), (1,2,2), (2,2,2)):
    print(f"igneous: factor must be 2,2,1 or 2,2,2. Got: {factor}")
    sys.exit()

  # comes from a converging infinite series proof
  constant = factor[0] * factor[1] * factor[2]
  if constant != 1:
    memory_bytes *= constant / (constant - 1)

  return memory_bytes

@main.command("view")
@click.argument("path", type=CloudPath())
@click.option('--browser/--no-browser', default=True, is_flag=True, help="Open the dataset in the system's default web browser.")
@click.option('--port', default=1337, help="localhost server port for the file server.", show_default=True)
@click.option('--ng', default="https://neuroglancer-demo.appspot.com/", help="Alternative Neuroglancer webpage to use.", show_default=True)
def view(path, browser, port, ng):
  """
  Open an on-disk dataset for viewing in neuroglancer.
  """
  # later improvements: 
  #   could use local neuroglancer
  #   modify the url to autopopulate params to avoid a click
  url = f"{ng}#!%7B%22layers%22:%5B%7B%22type%22:%22new%22%2C%22source%22:%22precomputed://http://localhost:{port}%22%2C%22tab%22:%22source%22%2C%22name%22:%22localhost:{port}%22%7D%5D%2C%22selectedLayer%22:%7B%22visible%22:true%2C%22layer%22:%22localhost:{port}%22%7D%2C%22layout%22:%224panel%22%7D"
  if browser:
    webbrowser.open(url, new=2)
  CloudVolume(path).viewer(port=port)



import pytest

import copy
from functools import partial
from collections import defaultdict
import os.path
import shutil

import numpy as np
from cloudvolume import CloudVolume, EmptyVolumeException, view
import cloudvolume.lib as lib
from cloudfiles import CloudFiles
from taskqueue import MockTaskQueue, TaskQueue
import tinybrain

import igneous
import igneous.task_creation as tc
from igneous.downsample_scales import create_downsample_scales

@pytest.fixture(scope="module", params=[ (512,512,128) ])
def transfer_data(request):
  data = np.zeros((512,512,128), dtype=np.uint8)
  i = 1
  for x in range(data.shape[0] // 64):
    for y in range(data.shape[1] // 64):
      for z in range(data.shape[2] // 64):
        data[64*x:64*(x+1), 64*y:64*(y+1), 64*z:64*(z+1)] = i
        i += 1 

  return data  

path_root = "/tmp/removeme/igneous/ccltask"
srcpath = f"file://{path_root}/src"
destpath = f"file://{path_root}/dest"

def rmdest():
  directory = f"{path_root}/dest"
  if os.path.exists(directory):
    shutil.rmtree(directory)
def rmsrc():
  directory = f"{path_root}/src"
  if os.path.exists(directory):
    shutil.rmtree(directory)

@pytest.fixture(scope="module")
def tq():
  return MockTaskQueue()

@pytest.fixture(scope="function")
def src_cv(transfer_data):
  rmsrc()
  rmdest()
  return CloudVolume.from_numpy(
    transfer_data, 
    vol_path=srcpath,
    resolution=(1,1,1), 
    voxel_offset=(0,0,0), 
    chunk_size=(64,64,64), 
    layer_type="image", 
    max_mip=0,
  )

def test_threshold_image():
  import igneous.tasks.image.ccl as ccl
  sz = 100
  image = np.arange(0,sz**3).reshape((sz,sz,sz), order="F")
  image = image.astype(np.uint32)

  res = ccl.threshold_image(image, None, None)  
  assert np.all(image == res)

  res = ccl.threshold_image(image, sz**3+1, None)  
  assert np.all(res == 1)

  res = ccl.threshold_image(image, None, 0)  
  assert np.all(res == 1)

  res = ccl.threshold_image(image, sz**3+1, 0)
  assert np.all(res == 1)

  res = ccl.threshold_image(image, None, 1)
  assert np.all(res[0,0,0] == 0)
  res[0,0,0] = 1
  assert np.all(res == 1)

  res = ccl.threshold_image(image, sz**3+1, 1)
  assert np.all(res[0,0,0] == 0)
  res[0,0,0] = 1
  assert np.all(res == 1)

@pytest.mark.parametrize("upper", (None,255,0))
@pytest.mark.parametrize("fill_missing", [True,False])
@pytest.mark.parametrize("dust_threshold", [0, (64**3) + 1])
def test_ccl_tasks(
  tq, src_cv, transfer_data, 
  upper, fill_missing,
  dust_threshold
):
  shape = (128,128,128)
  tasks = tc.create_ccl_face_tasks(
    src_cv.cloudpath, mip=0, shape=shape, 
    threshold_lte=upper, fill_missing=fill_missing,
    dust_threshold=dust_threshold,
  )
  tq.insert_all(tasks)

  cf = CloudFiles(src_cv.cloudpath)
  faces = list(cf.list(cf.join(src_cv.key, "ccl", "faces")))

  answer = [
    '1_1_1/ccl/faces/0-0-0-xy.cpso', '1_1_1/ccl/faces/0-0-0-xz.cpso', '1_1_1/ccl/faces/0-0-0-yz.cpso', 
    '1_1_1/ccl/faces/0-1-0-xy.cpso', '1_1_1/ccl/faces/0-1-0-xz.cpso', '1_1_1/ccl/faces/0-1-0-yz.cpso', 
    '1_1_1/ccl/faces/0-2-0-xy.cpso', '1_1_1/ccl/faces/0-2-0-xz.cpso', '1_1_1/ccl/faces/0-2-0-yz.cpso', 
    '1_1_1/ccl/faces/0-3-0-xy.cpso', '1_1_1/ccl/faces/0-3-0-xz.cpso', '1_1_1/ccl/faces/0-3-0-yz.cpso', 
    '1_1_1/ccl/faces/1-0-0-xy.cpso', '1_1_1/ccl/faces/1-0-0-xz.cpso', '1_1_1/ccl/faces/1-0-0-yz.cpso', 
    '1_1_1/ccl/faces/1-1-0-xy.cpso', '1_1_1/ccl/faces/1-1-0-xz.cpso', '1_1_1/ccl/faces/1-1-0-yz.cpso', 
    '1_1_1/ccl/faces/1-2-0-xy.cpso', '1_1_1/ccl/faces/1-2-0-xz.cpso', '1_1_1/ccl/faces/1-2-0-yz.cpso', 
    '1_1_1/ccl/faces/1-3-0-xy.cpso', '1_1_1/ccl/faces/1-3-0-xz.cpso', '1_1_1/ccl/faces/1-3-0-yz.cpso', 
    '1_1_1/ccl/faces/2-0-0-xy.cpso', '1_1_1/ccl/faces/2-0-0-xz.cpso', '1_1_1/ccl/faces/2-0-0-yz.cpso', 
    '1_1_1/ccl/faces/2-1-0-xy.cpso', '1_1_1/ccl/faces/2-1-0-xz.cpso', '1_1_1/ccl/faces/2-1-0-yz.cpso', 
    '1_1_1/ccl/faces/2-2-0-xy.cpso', '1_1_1/ccl/faces/2-2-0-xz.cpso', '1_1_1/ccl/faces/2-2-0-yz.cpso', 
    '1_1_1/ccl/faces/2-3-0-xy.cpso', '1_1_1/ccl/faces/2-3-0-xz.cpso', '1_1_1/ccl/faces/2-3-0-yz.cpso', 
    '1_1_1/ccl/faces/3-0-0-xy.cpso', '1_1_1/ccl/faces/3-0-0-xz.cpso', '1_1_1/ccl/faces/3-0-0-yz.cpso', 
    '1_1_1/ccl/faces/3-1-0-xy.cpso', '1_1_1/ccl/faces/3-1-0-xz.cpso', '1_1_1/ccl/faces/3-1-0-yz.cpso', 
    '1_1_1/ccl/faces/3-2-0-xy.cpso', '1_1_1/ccl/faces/3-2-0-xz.cpso', '1_1_1/ccl/faces/3-2-0-yz.cpso', 
    '1_1_1/ccl/faces/3-3-0-xy.cpso', '1_1_1/ccl/faces/3-3-0-xz.cpso', '1_1_1/ccl/faces/3-3-0-yz.cpso'
  ]
  faces.sort()
  answer.sort()

  assert len(faces) == 16 * 3
  assert faces == answer

  tasks = tc.create_ccl_equivalence_tasks(
    src_cv.cloudpath, mip=0, shape=shape, 
    threshold_lte=upper, fill_missing=fill_missing,
    dust_threshold=dust_threshold,
  )
  tq.insert_all(tasks)

  cf = CloudFiles(src_cv.cloudpath)
  equivs = list(cf.list(cf.join(src_cv.key, "ccl", "equivalences")))

  answer = [
    '1_1_1/ccl/equivalences/0-0-0.json', '1_1_1/ccl/equivalences/0-1-0.json', '1_1_1/ccl/equivalences/0-2-0.json', 
    '1_1_1/ccl/equivalences/0-3-0.json', '1_1_1/ccl/equivalences/1-0-0.json', '1_1_1/ccl/equivalences/1-1-0.json', 
    '1_1_1/ccl/equivalences/1-2-0.json', '1_1_1/ccl/equivalences/1-3-0.json', '1_1_1/ccl/equivalences/2-0-0.json', 
    '1_1_1/ccl/equivalences/2-1-0.json', '1_1_1/ccl/equivalences/2-2-0.json', '1_1_1/ccl/equivalences/2-3-0.json', 
    '1_1_1/ccl/equivalences/3-0-0.json', '1_1_1/ccl/equivalences/3-1-0.json', '1_1_1/ccl/equivalences/3-2-0.json', 
    '1_1_1/ccl/equivalences/3-3-0.json'
  ]
  answer.sort()
  equivs.sort()
  assert equivs == answer

  import igneous.tasks.image.ccl
  igneous.tasks.image.ccl.create_relabeling(src_cv.cloudpath, mip=0, shape=shape)

  tasks = tc.create_ccl_relabel_tasks(
    src_cv.cloudpath, destpath, 
    mip=0, shape=shape,
    threshold_lte=upper,
    fill_missing=fill_missing,
    dust_threshold=dust_threshold,
  )
  tq.insert_all(tasks)

  cv_dest = CloudVolume(destpath)
  cc_labels = cv_dest[:][:,:,:,0]

  uniq = np.unique(cc_labels)
  if dust_threshold > 0:
    if upper in (None, 0):
      assert len(uniq) == 1
      assert uniq[0] == 0
    else:
      assert len(uniq) == 1
      assert uniq[0] == 1
  else:
    if upper is None:
      assert len(uniq) == 128
      assert np.all(uniq == np.arange(1,129))
    elif upper == 255:
      assert len(uniq) == 1
      assert uniq[0] == 1
    elif upper == 0:
      assert len(uniq) == 1
      assert uniq[0] == 0

  rmsrc()
  rmdest()




import pytest

import shutil
import numpy as np

import os

from cloudvolume import CloudVolume
from cloudfiles import CloudFiles

layer_path = '/tmp/removeme/'

def create_storage(layer_name='layer'):
    stor_path = os.path.join(layer_path, layer_name)
    return CloudFiles('file://' + stor_path)

def create_layer(size, offset, layer_type="image", layer_name='layer', dtype=None):

    default = lambda dt: dtype or dt

    if layer_type == "image":
        random_data = np.random.randint(255, size=size, dtype=default(np.uint8))
    elif layer_type == 'affinities':
        random_data = np.random.uniform(low=0, high=1, size=size).astype(default(np.float32))
    elif layer_type == "segmentation":
        random_data = np.random.randint(0xFFFFFF, size=size, dtype=np.uint32)
    else:
        high = np.array([0], dtype=default(np.uint32)) - 1
        random_data = np.random.randint(high[0], size=size, dtype=default(np.uint32))
    
    storage = create_storage(layer_name)

    CloudVolume.from_numpy(
        random_data, 
        vol_path='file://' + layer_path + '/' + layer_name,
        resolution=(1,1,1), voxel_offset=offset, 
        chunk_size=(64,64,64), layer_type=layer_type, 
        max_mip=0,
    )
        
    return storage, random_data

def delete_layer(path=layer_path):
    if os.path.exists(path):
        shutil.rmtree(path)  

    
    
    


import pytest

import copy
from functools import partial
from collections import defaultdict
import os.path
import shutil

import numpy as np
from cloudvolume import CloudVolume, EmptyVolumeException, view
import cloudvolume.lib as lib
from cloudfiles import CloudFiles
from taskqueue import MockTaskQueue, TaskQueue
import tinybrain

import igneous
import igneous.task_creation as tc
from igneous.downsample_scales import create_downsample_scales

CHECKER = False

@pytest.fixture(scope="module", params=[ (512,512,128), (600, 600, 200) ])
def transfer_data(request):
  if not CHECKER:
    data = np.random.randint(0xFF, size=request.param, dtype=np.uint8)
  else:
    data = np.zeros((512,512,128), dtype=np.uint8)
    i = 1
    for x in range(8):
      for y in range(8):
        for z in range(2):
          data[64*x:64*(x+1), 64*y:64*(y+1), 64*z:64*(z+1)] = i
          i += 1 

  ds_data = tinybrain.downsample_with_averaging(data, factor=[2, 2, 1, 1], num_mips=5)
  ds_data.insert(0, data)

  for i in range(len(ds_data)):
    while ds_data[i].ndim < 4:
      ds_data[i] = ds_data[i][..., np.newaxis]

  return ds_data  

path_root = "/tmp/removeme/transfertask"
srcpath = f"file://{path_root}/src"
destpath = f"file://{path_root}/dest"

def rmdest():
  directory = f"{path_root}/dest"
  if os.path.exists(directory):
    shutil.rmtree(directory)
def rmsrc():
  directory = f"{path_root}/src"
  if os.path.exists(directory):
    shutil.rmtree(directory)

@pytest.fixture(scope="module")
def tq():
  return MockTaskQueue()

@pytest.fixture(scope="function")
def src_cv(transfer_data):
  rmsrc()
  rmdest()
  return CloudVolume.from_numpy(
    transfer_data[0], 
    vol_path=srcpath,
    resolution=(1,1,1), 
    voxel_offset=(0,0,0), 
    chunk_size=(64,64,64), 
    layer_type="image", 
    max_mip=0,
  )

def test_transfer_task_vanilla(tq, src_cv, transfer_data):
  tasks = tc.create_transfer_tasks(src_cv.cloudpath, destpath)
  tq.insert_all(tasks)

  dest_cv = CloudVolume(destpath)
  assert np.all(src_cv[:] == dest_cv[:])
  rmsrc()
  rmdest()

def test_transfer_task_rechunk(tq, src_cv, transfer_data):
  tasks = tc.create_transfer_tasks(
    src_cv.cloudpath, destpath, chunk_size=(50,50,50)
  )
  tq.insert_all(tasks)
  dest_cv = CloudVolume(destpath)
  assert len(dest_cv.scales) == 5
  assert np.all(src_cv[:] == dest_cv[:])
  for mip in range(1, 5):
    dest_cv.mip = mip
    assert np.all(dest_cv[:] == transfer_data[mip])
  rmsrc()
  rmdest()

# chunk size (64,64,64) should test transfer_to pathway
@pytest.mark.parametrize("chunk_size", [ (50,50,50), (64,64,64) ])
def test_transfer_task_skip_downsample(tq, src_cv, chunk_size):
  tasks = tc.create_transfer_tasks(
    src_cv.cloudpath, destpath, 
    chunk_size=chunk_size, skip_downsamples=True
  )
  tq.insert_all(tasks)
  dest_cv = CloudVolume(destpath)
  assert len(dest_cv.scales) == 1
  assert np.all(src_cv[:] == dest_cv[:])
  rmsrc()
  rmdest()

def test_transfer_task_dest_offset(tq, src_cv, transfer_data):
  tasks = tc.create_transfer_tasks(
    src_cv.cloudpath, destpath, 
    chunk_size=(50,50,50), dest_voxel_offset=(100, 100, 100)
  )
  tq.insert_all(tasks)
  dest_cv = CloudVolume(destpath)
  assert len(dest_cv.scales) == 5
  assert tuple(dest_cv.voxel_offset) == (100, 100, 100)
  assert tuple(src_cv.voxel_offset) == (0, 0, 0)
  assert np.all(src_cv[:] == dest_cv[:])
  for mip in range(1, 5):
    dest_cv.mip = mip
    assert np.all(dest_cv[:] == transfer_data[mip])
  rmsrc()
  rmdest()

def test_transfer_task_subset(tq, src_cv, transfer_data):
  dest_cv = CloudVolume(destpath, info=copy.deepcopy(src_cv.info))
  dest_cv.scale["size"] = [256, 256, 64]
  dest_cv.commit_info()

  tasks = tc.create_transfer_tasks(
    src_cv.cloudpath, destpath, 
    chunk_size=(64,64,64), 
    translate=(-128, -128, -64),
  )
  tq.insert_all(tasks)
  dest_cv.refresh_info()

  assert len(dest_cv.scales) == 3
  assert np.all(src_cv[128:128+256, 128:128+256, 64:64+64] == dest_cv[:])

  rmsrc()
  rmdest()




"""
Perform connected components labeling on the image.

Result will be a 6-connected labeling of the input
image. All steps must use the same task shape. 

As each task uses a label offset to differentiate its
CCL labels from adjacent tasks, the largest possible
image that can be handled has 2^64 voxels which is
order of magnitude about size of a whole mouse brain.

Each of the steps are labeled with their sequence number.
Their order is:
  (1) Generate 3 back faces for each task with 
    1 voxel overlap (so they can be referenced by 
    adjacent tasks). [ CCLFacesTask ]
  (2) Compute linkages between CCL tasks and save the results 
    in a database. [ CCLEquivalancesTask ]
  (3) Compute a global union find from the linkage 
    data and from that a global relabeling scheme which 
    is saved in the database. [ create_relabeling ]
  (4) Apply the relabeling scheme to the image. [ RelabelCCLTask ]
"""
from typing import Optional, Union
from collections import defaultdict

import cc3d
import fastremap
import numpy as np
import compresso

from tqdm import tqdm

from cloudfiles import CloudFiles
from taskqueue import queueable

from cloudvolume import CloudVolume, Bbox, Vec
from cloudvolume.lib import sip

from ...types import ShapeType

__all__ = [
  "CCLFacesTask",
  "CCLEquivalancesTask",
  "RelabelCCLTask",
]

class DisjointSet:
  def __init__(self):
    self.data = {} 
  def makeset(self, x):
    self.data[x] = x
    return x
  def find(self, x):
    if not x in self.data:
      return None
    i = self.data[x]
    while i != self.data[i]:
      self.data[i] = self.data[self.data[i]]
      i = self.data[i]
    return i
  def union(self, x, y):
    i = self.find(x)
    j = self.find(y)
    if i is None:
      i = self.makeset(x)
    if j is None:
      j = self.makeset(y)

    if i < j:
      self.data[j] = i
    else:
      self.data[i] = j

def compute_task_number(grid_size, gridpoint) -> int:
  return int(
    gridpoint.x + grid_size.x * (
      gridpoint.y + grid_size.y * gridpoint.z
    )
  )

def compute_label_offset(shape, grid_size, gridpoint) -> int:
  # a task sequence number counting from 0 according to
  # where the task is located in space (so we can recompute 
  # it in the second pass easily)
  task_num = compute_task_number(grid_size, gridpoint)
  return task_num * shape.x * shape.y * shape.z

def threshold_image(
  image:np.ndarray, 
  threshold_lte:Optional[Union[int,float]], 
  threshold_gte:Optional[Union[int,float]]
) -> np.ndarray:
  if threshold_gte is None and threshold_lte is None:
    return image

  thresholded = np.zeros(image.shape, dtype=np.uint8, order="F")
  if threshold_gte is not None:
    thresholded += image >= threshold_gte
  if threshold_lte is not None:
    if threshold_gte is not None:
      thresholded *= image <= threshold_lte
    else:
      thresholded += image <= threshold_lte
  return thresholded

def blackout_non_face_rails(
  labels:np.ndarray, shape:ShapeType
) -> np.ndarray:
  """
  For 6-connectivity, we need to black out the
  "rails" that would represent a higher connectivity
  to ensure that the labels are actually present
  when computing the relabeling later.
  """
  slcs = [
    np.s_[shape[0],shape[1],:],
    np.s_[shape[0],:,shape[2]],
    np.s_[:,shape[1],shape[2]]
  ]

  for slc in slcs:
    try:
      labels[slc] = 0
    except IndexError:
      pass

  return labels

@queueable
def CCLFacesTask(
  cloudpath:str, mip:int, 
  shape:ShapeType, offset:ShapeType,
  threshold_gte:Optional[Union[float,int]] = None,
  threshold_lte:Optional[Union[float,int]] = None,
  fill_missing:bool = False,
  dust_threshold:int = 0,
):
  """
  (1) Generate x,y,z back faces of each 1vx overlap task
  as compresso encoded 2D images.

  For continuous data, greater than or equal to (gte) 
  or less than or equal to (lte) thresholds 
  can be applied either individually or together. If neither
  is specified, no modification of the input will occur.

  These images are stored in e.g. 32_32_40/ccl/faces/ and have
  the following scheme where the numbers are the gridpoint
  location and the letters indicate which face plane.
    1-2-0-xy.cpso
    1-2-0-xz.cpso
    1-2-0-yz.cpso
  """
  shape = Vec(*shape)
  offset = Vec(*offset)
  bounds = Bbox(offset, offset + shape + 1) # 1 vx overlap

  if bounds.subvoxel():
    return

  cv = CloudVolume(cloudpath, mip=mip, fill_missing=fill_missing)
  bounds = Bbox.clamp(bounds, cv.meta.bounds(mip))

  grid_size = np.ceil((cv.bounds / shape).size3())
  gridpoint = np.floor(bounds.center() / shape).astype(int)
  label_offset = compute_label_offset(shape + 1, grid_size, gridpoint)
  
  labels = threshold_image(cv[bounds][...,0], threshold_lte, threshold_gte)
  labels = blackout_non_face_rails(labels, shape)
  if dust_threshold > 0:
    labels = cc3d.dust(
      labels, threshold=dust_threshold, 
      connectivity=6, in_place=True
    )
  cc_labels = cc3d.connected_components(labels, connectivity=6, out_dtype=np.uint64)
  cc_labels += label_offset
  cc_labels[labels == 0] = 0

  # Uploads leading faces for adjacent tasks to examine
  slices = [
    cc_labels[:,:,-1],
    cc_labels[:,-1,:],
    cc_labels[-1,:,:],
  ]
  slices = [ compresso.compress(slc) for slc in slices ]
  filenames = [
    f'{gridpoint.x}-{gridpoint.y}-{gridpoint.z}-xy.cpso',
    f'{gridpoint.x}-{gridpoint.y}-{gridpoint.z}-xz.cpso',
    f'{gridpoint.x}-{gridpoint.y}-{gridpoint.z}-yz.cpso'
  ]

  cf = CloudFiles(cloudpath)
  filenames = [
    cf.join(cv.key, 'ccl', 'faces', fname) for fname in filenames
  ]
  cf.puts(zip(filenames, slices), compress='br')

@queueable
def CCLEquivalancesTask(
  cloudpath:str, mip:int,
  shape:ShapeType, offset:ShapeType,
  threshold_gte:Optional[Union[float,int]] = None,
  threshold_lte:Optional[Union[float,int]] = None,
  fill_missing:bool = False,
  dust_threshold:int = 0,
):
  """
  (2) Generate linkages between tasks by comparing the 
  front face of the task with the back faces of the 
  three adjacent tasks saved from the first step.

  Writes output to e.g. 32_32_40/ccl/equivalences/
  """
  shape = Vec(*shape)
  offset = Vec(*offset)
  bounds = Bbox(offset, offset + shape + 1) # 1 vx overlap

  if bounds.subvoxel():
    return

  cv = CloudVolume(cloudpath, mip=mip, fill_missing=fill_missing)
  bounds = Bbox.clamp(bounds, cv.meta.bounds(mip))

  grid_size = np.ceil((cv.bounds / shape).size3())
  gridpoint = np.floor(bounds.center() / shape).astype(int)
  label_offset = compute_label_offset(shape + 1, grid_size, gridpoint)
  
  equivalences = DisjointSet()

  labels = threshold_image(cv[bounds][...,0], threshold_lte, threshold_gte)
  labels = blackout_non_face_rails(labels, shape)
  if dust_threshold > 0:
    labels = cc3d.dust(
      labels, threshold=dust_threshold, 
      connectivity=6, in_place=True
    )
  cc_labels, N = cc3d.connected_components(
    labels, connectivity=6, 
    out_dtype=np.uint64, return_N=True
  )
  cc_labels += label_offset
  cc_labels[labels == 0] = 0

  for i in range(1, N+1):
    equivalences.makeset(i + label_offset)

  cf = CloudFiles(cloudpath)
  filenames = [
    f'{gridpoint.x}-{gridpoint.y}-{gridpoint.z-1}-xy.cpso',
    f'{gridpoint.x}-{gridpoint.y-1}-{gridpoint.z}-xz.cpso',
    f'{gridpoint.x-1}-{gridpoint.y}-{gridpoint.z}-yz.cpso'
  ]
  filenames = [
    cf.join(cv.key, 'ccl', 'faces', fname) for fname in filenames
  ]

  slices = cf.get(filenames, return_dict=True)
  for key in slices:
    if slices[key] is None:
      continue
    face = compresso.decompress(slices[key])
    if '-xy' in key:
      face = face[:shape.x,:shape.y,0]
    elif '-xz' in key:
      face = face[:shape.x,:shape.z,0]
    else:
      face = face[:shape.y,:shape.z,0]
    slices[key] = face

  prev = [ slices[filenames[i]] for i in (0,1,2) ]

  cur = [
    cc_labels[:shape.x,:shape.y,0],
    cc_labels[:shape.x,0,:shape.z],
    cc_labels[0,:shape.y,:shape.z],
  ]

  for prev_i, cur_i in zip(prev, cur):
    if prev_i is None:
      continue

    mapping = fastremap.inverse_component_map(cur_i, prev_i)
    for task_label, adj_labels in mapping.items():
      if task_label == 0:
        continue
      for adj_label in fastremap.unique(adj_labels):
        if adj_label != 0:
          equivalences.union(int(task_label), int(adj_label))

  cf = CloudFiles(cloudpath)
  out_name = cf.join(cv.key, 'ccl', 'equivalences', f'{gridpoint.x}-{gridpoint.y}-{gridpoint.z}.json')
  cf.put_json(
    out_name, 
    { str(k): int(v) for k,v in equivalences.data.items() }, 
    compress='br'
  )

@queueable
def RelabelCCLTask(
  src_path:str, dest_path:str, mip:int,
  shape:ShapeType, offset:ShapeType,
  threshold_gte:Optional[Union[float,int]] = None,
  threshold_lte:Optional[Union[float,int]] = None,
  fill_missing:bool = False,
  dust_threshold:int = 0,
):
  """
  (4) Retrieves the relabeling for this task from the
  database, applies it, and saves the resulting image
  to the destination path. Upon saving, the 1 voxel
  overlap is omitted.
  """
  shape = Vec(*shape)
  offset = Vec(*offset)
  bounds = Bbox(offset, offset + shape + 1) # 1 vx overlap

  if bounds.subvoxel():
    return

  cv = CloudVolume(src_path, mip=mip, fill_missing=fill_missing)
  bounds = Bbox.clamp(bounds, cv.meta.bounds(mip))

  grid_size = np.ceil((cv.bounds / shape).size3())
  gridpoint = np.floor(bounds.center() / shape).astype(int)
  label_offset = compute_label_offset(shape + 1, grid_size, gridpoint)
  task_num = compute_task_number(grid_size, gridpoint)
  
  cf = CloudFiles(src_path)
  mapping_path = cf.join(cv.key, "ccl", "relabel", f"{task_num}.json")
  mapping = cf.get_json(mapping_path) or {}
  mapping =  { int(k):int(v) for k,v in mapping.items() }
  mapping[0] = 0

  labels = threshold_image(cv[bounds][...,0], threshold_lte, threshold_gte)
  labels = blackout_non_face_rails(labels, shape)
  if dust_threshold > 0:
    labels = cc3d.dust(
      labels, threshold=dust_threshold, 
      connectivity=6, in_place=True
    )
  cc_labels, N = cc3d.connected_components(
    labels, connectivity=6, 
    out_dtype=np.uint64, return_N=True
  )
  cc_labels += label_offset
  cc_labels[labels == 0] = 0

  task_voxels = shape.x * shape.y * shape.z
  fastremap.remap(cc_labels, mapping, in_place=True)

  # Final upload without overlap
  dest_cv = CloudVolume(dest_path, mip=mip)
  bounds = Bbox(offset, offset + shape)
  bounds = Bbox.clamp(bounds, dest_cv.meta.bounds(mip))
  shape = bounds.size3()

  cc_labels = cc_labels[:shape.x,:shape.y,:shape.z].astype(dest_cv.dtype)
  cc_labels = cc_labels[:,:,:,np.newaxis]
  dest_cv[bounds] = cc_labels

def create_relabeling(cloudpath, mip, shape):
  """
  (3) Computes a relabeling from the linkages saved 
  from (2) and then saves them.

  Writes output to e.g. 32_32_40/ccl/relabel/
  and also .../ccl/max_label.json which contains
  the largest label.
  """
  cv = CloudVolume(cloudpath, mip=mip)
  cf = CloudFiles(cloudpath)
  all_eqpaths = cf.list(cf.join(cv.key, "ccl", "equivalences"))

  equivalences = DisjointSet()

  with tqdm(desc="Creating Union-Find", total=0) as pbar:
    for eqpaths in sip(all_eqpaths, 5000):
      eqdicts = cf.get_json(eqpaths)
      pbar.total += sum(( len(datum) for datum in eqdicts ))
      pbar.refresh()
      
      for data in eqdicts:
        for val1, val2 in data.items():
          equivalences.union(int(val1), int(val2))
          pbar.update()

      del eqdicts

  relabel = {}
  next_label = 1
  for key in tqdm(equivalences.data.keys(), desc="Renumbering"):
    lbl = equivalences.find(key)
    if lbl not in relabel:
      relabel[key] = next_label
      relabel[lbl] = next_label
      next_label += 1
    else:
      relabel[key] = relabel[lbl]

  del equivalences

  max_label_fname = cf.join(cv.key, "ccl", "max_label.json")
  cf.put_json(max_label_fname, [ next_label - 1 ])

  task_size = Vec(*shape) + 1
  task_voxels = task_size.x * task_size.y * task_size.z

  buckets = defaultdict(dict)
  for before_val, after_val in relabel.items():
    task_num = int(before_val // task_voxels)
    buckets[task_num][before_val] = after_val

  del relabel

  cf.put_jsons(
    (
      (cf.join(cv.key, "ccl", "relabel", f"{task_num}.json"), relabeling) 
      for task_num, relabeling in buckets.items()
    ),
    total=len(buckets),
    compress="br",
    progress=True
  )

def clean_intermediate_files(src, mip):
  cv = CloudVolume(src, mip)
  cf = CloudFiles(src)
  cf.delete(cf.list(cf.join(cv.key, "ccl")))





